from .generators import AttributeGenerator, TagGenerator
from .generators.templates import ChoiceTemplate, RandomTemplate
from .generators.templates.base import BaseTemplate

__all__ = [
    "AttributeGenerator",
    "TagGenerator",
    "ChoiceTemplate",
    "RandomTemplate",
    "BaseTemplate",
]



"""
Module for combinators defined in css.

They they combine other selectors in a way that gives them a useful relationship
to each other and the location of content in the document.

Soupsavvy provides combinators that are used to combine multiple SoupSelector
objects in a similar fashion to CSS combinators.

Classes
-------
ChildCombinator - equivalent of CSS child combinator (>)
NextSiblingCombinator - equivalent of CSS adjacent sibling combinator (+)
SubsequentSiblingCombinator - equivalent of CSS general sibling combinator (*)
DescentCombinator - equivalent of CSS descendant combinator (" ")
SelectorList - equivalent of CSS selector list (,) or :is() pseudo-class
OrSelector - alias for SelectorList
ParentCombinator - matches parent of preceding selector
AncestorCombinator - matches ancestor of preceding selector

Notes
-----
For more information on CSS combinators see:
https://developer.mozilla.org/en-US/docs/Learn/CSS/Building_blocks/Selectors/Combinators
"""

from abc import abstractmethod
from functools import reduce
from typing import Optional, Type

from bs4 import Tag

from soupsavvy.selectors.base import CompositeSoupSelector, SoupSelector
from soupsavvy.selectors.relative import (
    RelativeAncestor,
    RelativeChild,
    RelativeDescendant,
    RelativeNextSibling,
    RelativeParent,
    RelativeSelector,
    RelativeSubsequentSibling,
)
from soupsavvy.selectors.tag_utils import TagIterator, TagResultSet


class BaseCombinator(CompositeSoupSelector):
    # order of selectors is relevant in context of results
    COMMUTATIVE = False

    def __init__(
        self,
        selector1: SoupSelector,
        selector2: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes Combinator object with provided positional arguments.
        At least two SoupSelector object are required to create Combinator.

        Parameters
        ----------
        selectors: SoupSelector
            SoupSelector objects to match accepted as positional arguments.

        Notes
        -----
        Object can be initialized with more than two SoupSelector objects,
        which would be equal to chaining multiple combinators of the same type.

        For example, chaining child combinator in css:

        Example
        -------
        >>> div > a > span

        translated to soupsavvy would be:

        Example
        -------
        >>> ChildCombinator(TypeSelector("div"), TypeSelector("a"), TypeSelector("span"))

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of SoupSelector.
        """
        super().__init__([selector1, selector2, *selectors])

    @property
    @abstractmethod
    def _selector(self) -> Type[RelativeSelector]:
        """
        Returns type of the relative selector that is used
        to perform a single step search in the combinator selector.

        Returns
        -------
        Type[RelativeSelector]
            Type of the relative selector that is used in the combinator.
            Selector instance of this type is initialized
            with each step in the combinator.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} is a base class "
            "and does not implement '_selector' property."
        )

    def _find_first_step(
        self, step: SoupSelector, tag: Tag, recursive: bool
    ) -> TagResultSet:
        """
        Returns results of the first step in the combinator selector,
        given Tag object that is being searched and recursive behavior
        passed to find method by user.

        Parameters
        ----------
        tag: Tag
            Tag object that's being searched by first step in the combinator.
        recursive: bool
            Recursive behavior passed to find method by user.

        Returns
        -------
        TagResultSet
            Results of the first step in the combinator selector.
        """
        return TagResultSet(step.find_all(tag, recursive=recursive))

    def _order_results(
        self, results: TagResultSet, tag: Tag, recursive: bool
    ) -> TagResultSet:
        """
        Orders results of find_all method of the combinator selector, given
        initial Tag object that was passed to find_all method and recursive behavior.

        Parameters
        ----------
        results: TagResultSet
            Results of the combinator selector.
        tag: Tag
            Initial Tag object that was passed to find_all method.
        recursive: bool
            Recursive behavior passed to find method by user.

        Returns
        -------
        TagResultSet
            Ordered results of the combinator selector.
        """
        return TagResultSet(list(TagIterator(tag, recursive=True))) & results

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        results = TagResultSet()

        for i, step in enumerate(self.selectors):
            if i == 0:
                results |= self._find_first_step(
                    step=step, tag=tag, recursive=recursive
                )
                continue

            if not results:
                break

            selector = self._selector(step)
            results = TagResultSet(
                reduce(
                    list.__add__,
                    # each relative selector has defined recursive behavior
                    (selector.find_all(element) for element in results.fetch()),
                )
            )

        results = self._order_results(results=results, tag=tag, recursive=recursive)
        return results.fetch(limit)


class BaseAncestorCombinator(BaseCombinator):
    """
    Base class for ancestor combinators, that are specific type of combinators,
    unlike other combinators, they move up the tree of elements, rather than down
    after finding first step.

    * Elements that match first step can be found
    anywhere in the tree, regardless of `recursive` parameter.
    * Final results should contain only children if `recursive` is False.
    """

    def _find_first_step(
        self, step: SoupSelector, tag: Tag, recursive: bool
    ) -> TagResultSet:
        # always look for all descendants in first step
        return TagResultSet(step.find_all(tag, recursive=True))

    def _order_results(
        self, results: TagResultSet, tag: Tag, recursive: bool
    ) -> TagResultSet:
        # respect recursive parameter while ordering results
        return TagResultSet(list(TagIterator(tag, recursive=recursive))) & results


class ChildCombinator(BaseCombinator):
    """
    Class representing a child combinator in CSS selectors.

    Example
    -------
    >>> ChildCombinator(TypeSelector("div"), TypeSelector("a"))

    matches all 'a' elements that are direct children of 'div' elements.

    Example
    -------
    >>> <div class="widget"><a>Hello World</a></div> ✔️
    >>> <div class="widget"><span></span><a>Hello World</a></div> ✔️
    >>> <span class="widget"><a>Hello World</a></span> ❌
    >>> <div class="menu"><span>Hello World</span></div> ❌

    Object can be created as well by using greater than operator '>'
    on two SoupSelector objects.

    Example
    -------
    >>> TypeSelector("div") > TypeSelector("a")

    Which is equivalent to the first example.

    This is an equivalent of CSS child combinator, that matches elements
    that are direct children of a specified element.

    Example
    -------
    >>> .widget > a { color: red; }

    which translated to soupsavvy would be:

    Example
    -------
    >>> ChildCombinator(
    ...     AttributeSelector("class", "widget"),
    ...     TypeSelector("a")
    ... )

    or with use of `>` operator:

    Example
    -------
    >>> AttributeSelector("class", "widget") > TypeSelector("a")

    Notes
    -----
    For more information on child combinator see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/Child_combinator
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeChild


class NextSiblingCombinator(BaseCombinator):
    """
    Class representing a next sibling combinator in CSS selectors.

    Example
    -------
    >>> NextSiblingCombinator(TypeSelector("div"), TypeSelector("a"))

    matches all 'a' elements that immediately follow 'div' elements,
    it means that both elements are children of the same parent element.

    Example
    -------
    >>> <div class="widget"></div><a>Hello World</a> ✔️
    >>> <div class="widget"><a>Hello World</a></div> ❌
    >>> <div class="widget"></div><span></span><a>Hello World</a> ❌

    Object can be created as well by using plus operator '+'
    on two SoupSelector objects.

    Example
    -------
    >>> div + a { color: red; }

    Which translated to soupsavvy would be:

    Example
    -------
    >>> NextSiblingCombinator(TypeSelector("div"), TypeSelector("a"))
    >>> TypeSelector("div") + TypeSelector("a")

    Notes
    -----
    For more information on next sibling combinator see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/Next-sibling_combinator

    This is also known as the adjacent sibling combinator.
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeNextSibling


class SubsequentSiblingCombinator(BaseCombinator):
    """
    Class representing a subsequent sibling combinator in CSS selectors.
    Subsequent sibling combinator separates two selectors
    and matches all instances of the second element that follow the first element
    (not necessarily immediately) and share the same parent element.

    Example
    -------
    >>> SubsequentSiblingCombinator(TypeSelector("div"), TypeSelector("a"))

    matches all 'a' elements that follow 'div' elements.

    Example
    -------
    >>> <div class="widget"></div><a>Hello World</a> ✔️
    >>> <div class="widget"><span></span><a>Hello World</a></div> ✔️
    >>> <span class="widget"><a>Hello World</a></span> ❌
    >>> <a>Hello World</a><div class="menu"></div> ❌

    Object can be created as well by using multiplication operator '*'
    on two SoupSelector objects, due to the lack of support for '~' operator
    between two operands.
    Example
    -------
    >>> div ~ a { color: red; }

    Which translated to soupsavvy would be:

    Example
    -------
    >>> TypeSelector("div") * TypeSelector("a")
    >>> SubsequentSiblingCombinator(TypeSelector("div"), TypeSelector("a"))

    Notes
    -----
    For more information on subsequent sibling combinator see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/Subsequent-sibling_combinator

    This combinator is also known as `general sibling combinator`
    or `adjacent sibling combinator`.
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeSubsequentSibling


class DescendantCombinator(BaseCombinator):
    """
    Class representing a descent combinator in CSS selectors.
    Descent combinator separates two selectors and matches all instances
    of the second element that are descendants of the first element.

    Two SoupSelector objects are required to create DescentCombinator,
    but more can be provided as positional arguments, which binds them
    in a sequence of steps to match.

    Example
    -------
    >>> DescentCombinator(
    >>>     TypeSelector("div"),
    >>>     AttributeSelector(name="href", value="google.com")
    >>> )

    matches all descendants of 'div' element
    that have href attribute equal to 'google.com'.

    Example
    -------
    >>> <div><a href="google.com"></a></div> ✔️
    >>> <div><div><a href="google.com"></a></div></div> ✔️
    >>> <div><a href="duckduckgo.com"></a></div> ❌
    >>> <span><a href="google.com"></a></span> ❌
    >>> <a href="google.com"></a> ❌
    >>> <div class="widget"></div> ❌

    Object can be created as well by using right shift operator '>>'
    on two SoupSelector objects.

    Example
    -------
    >>> div a { color: red; }

    Which translated to soupsavvy would be:

    Example
    -------
    >>> TypeSelector("div") >> TypeSelector("a")
    >>> DescentCombinator(TypeSelector("div"), TypeSelector("a"))

    Notes
    -----
    For more information on subsequent sibling combinator see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/Descendant_combinator
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeDescendant


class ParentCombinator(BaseAncestorCombinator):
    """
    Parent combinator separates two selectors and matches all instances
    of the second element that are parents of the first element.

    Two `SoupSelector` objects are required to create ParentCombinator,
    but more can be provided as positional arguments, which binds them
    in a sequence of steps to match.

    Example
    -------
    >>> ParentCombinator(
    ...     TypeSelector("a"),
    ...     AttributeSelector("class", "menu"),
    ...     TypeSelector("div"),
    ... )

    The given selector would first look for 'a' elements, then find elements with
    'menu' class that are parents of 'a' elements, and finally find 'div' elements
    that are parents of 'menu' elements.

    Example
    -------
    >>> <div><span class="menu"><a href="/shop"></a></span></div> ✔️
    >>> <div><span class="menu"><div><a href="/shop"></a></div></span></div> ❌
    >>> <div><a href="/shop"></a><span class="menu"></span></div> ❌
    >>> <div><a href="/shop"></a></span></div> ❌

    It is equivalent to using `HasSelector` with child combinator and narrowing
    down results to `div` types with `TypeSelector`.

    Example
    -------
    >>> HasSelector(
    ...     Anchor > AttributeSelector("class", "menu") > TypeSelector("a")
    ... ) & TypeSelector("div")

    Although this combinator does not have its counterpart in CSS, it can be
    represented as:

    Example
    -------
    >>> div:has(> .menu > a)

    Object can be created as well by using lt operator '<'
    on two SoupSelector objects.

    Example
    -------
    >>> div:has(> a) { color: red; }

    Which translated to soupsavvy would be:

    Example
    -------
    >>> TypeSelector("a") < TypeSelector("div")
    >>> ParentCombinator(TypeSelector("a"), TypeSelector("div"))
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeParent


class AncestorCombinator(BaseAncestorCombinator):
    """
    Ancestor combinator separates two selectors and matches all instances
    of the second element that are ancestors of the first element.

    Two `SoupSelector` objects are required to create AncestorCombinator,
    but more can be provided as positional arguments, which binds them
    in a sequence of steps to match.

    Example
    -------
    >>> AncestorCombinator(
    ...     TypeSelector("a"),
    ...     AttributeSelector("class", "menu"),
    ...     TypeSelector("div"),
    ... )

    The given selector would first look for 'a' elements, then find elements with
    'menu' class that are ancestors of 'a' elements, and finally find 'div' elements
    that are ancestors of 'menu' elements.

    Example
    -------
    >>> <div><span class="menu"><a href="/shop"></a></span></div> ✔️
    >>> <div><span class="menu"><div><a href="/shop"></a></div></span></div> ✔️
    >>> <div><a href="/shop"></a><span class="menu"></span></div> ❌
    >>> <div><a href="/shop"></a></span></div> ❌

    It is equivalent to using `HasSelector` with (default) descendant combinator
    and narrowing down results to `div` types with `TypeSelector`.

    Example
    -------
    >>> HasSelector(
    ...     AttributeSelector("class", "menu") > TypeSelector("a")
    ... ) & TypeSelector("div")

    Although this combinator does not have its counterpart in CSS, it can be
    represented as has selector, where descendant combinator is implied:

    Example
    -------
    >>> div:has(.menu a)

    Object can be created as well by using left shift operator '<<'
    on two SoupSelector objects.

    Example
    -------
    >>> div:has(a) { color: red; }

    Which translated to soupsavvy would be:

    Example
    -------
    >>> TypeSelector("a") << TypeSelector("div")
    >>> AncestorCombinator(TypeSelector("a"), TypeSelector("div"))
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeAncestor


class SelectorList(CompositeSoupSelector):
    """
    Class representing a list of selectors in CSS,
    a selector list is a comma-separated list of selectors,
    that selects all the matching nodes.

    Class represents an Union of multiple soup selectors.
    Provides elements matching any of the selectors in an Union.

    At least two SoupSelector objects are required to create SelectorList,
    but more can be provided as positional arguments.

    Example
    -------
    >>> SelectorList(
    >>>     TypeSelector("a"),
    >>>     TypeSelector("div", [AttributeSelector(name="class", value="widget")])
    >>> )

    matches all elements that have "a" tag name OR 'class' attribute "widget".

    Example
    -------
    >>> <a>Hello World</a> ✔️
    >>> <div class="widget">Hello World</div> ✔️
    >>> <div>Hello Python</div> ❌

    This is an equivalent of CSS comma selector (selector list) or :is() selector.

    Example
    -------
    >>> h1, h1 { color: red;}
    >>> :is(h1, h2) { color: red; }

    This example translated to soupsavvy would be:

    Example
    -------
    >>> SoupUnionTag(TypeSelector("h1"), TypeSelector("h2"))
    >>> TypeSelector("h1") | TypeSelector("h2")

    Object can be created as well by using bitwise OR operator '|'
    on two SoupSelector objects.

    Example
    -------
    >>> TypeSelector("a") | TypeSelector("div", [AttributeSelector(name="class", value="widget")])

    Which is equivalent to the first example.

    Notes
    -----
    This Combinator does not inherit from BaseCombinator,
    as its logic differs from other combinators.

    For more information on selector list see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/Selector_list
    """

    def __init__(
        self,
        selector1: SoupSelector,
        selector2: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes SelectorList object with provided positional arguments.
        At least two SoupSelector objects are required to create SelectorList.

        Parameters
        ----------
        selectors: SoupSelector
            SoupSelector objects to match accepted as positional arguments.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of SoupSelector.
        """
        super().__init__([selector1, selector2, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        results = TagResultSet()

        for selector in self.selectors:
            results |= TagResultSet(
                selector.find_all(tag, recursive=recursive),
            )

        # keep order of tags and limit
        results = TagResultSet(list(TagIterator(tag, recursive=recursive))) & results
        return results.fetch(limit)



"""
Module with `soupsavvy` interfaces used across the package.

* `Executable` - Interface for operations that can be executed on single argument.
* `Comparable` - Interface for objects that can be compared for equality.
* `TagSearcher` - Interface for objects that can search within BeautifulSoup tags.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Optional

from bs4 import Tag

import soupsavvy.exceptions as exc


class Executable(ABC):
    """
    Interface for operations that can be executed on single argument.
    Derived classes must implement the `execute` method.
    """

    @abstractmethod
    def execute(self, arg: Any) -> Any:
        """Executes the operation on the given argument."""
        pass


class Comparable(ABC):
    """
    Interface for objects that can be compared for equality.
    Derived classes must implement the `__eq__` method.
    """

    @abstractmethod
    def __eq__(self, x: Any) -> bool:
        pass


# possible exceptions raised when TagSearcher fails
TagSearcherExceptions = (exc.FailedOperationExecution, exc.TagNotFoundException)


class TagSearcher(ABC):
    """
    Interface for objects that can search within BeautifulSoup tags.
    """

    @abstractmethod
    def find(
        self,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Any:
        """
        Finds first target information within the given tag and returns the result.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object.
        strict : bool, optional
            If True, enforces results to be found in the tag, by default False.
        recursive : bool, optional
            bs4.Tag.find method parameter that specifies if search should be recursive.
            If set to False, only direct children of the tag will be searched.
            By default True.

        Returns
        -------
        Any
            Result of the search operation.
        """

    @abstractmethod
    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Any]:
        """
        Finds all target information within the given tag and returns the result.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object.
        recursive : bool, optional
            bs4.Tag.find method parameter that specifies if search should be recursive.
            If set to False, only direct children of the tag will be searched.
            By default True.
        limit : int, optional
            bs4.Tag.find_all method parameter that specifies maximum number of results
            to return. If set to None, all results are returned. By default None.

        Returns
        -------
        list[Any]
            A list of results, if none found, the list is empty.
        """



"""
Module for the BaseGenerator abstract class that is a parent for all generators.

soupsavvy generator is anything that generates content,
they should all inherit from this class and implement the 'generate' method.
"""

from abc import ABC, abstractmethod


class BaseGenerator(ABC):
    """
    Abstract base class for all kind of soupsavvy generators.

    All components generating content should inherit from this class
    and implement the 'generate' method to return the generated content.
    """

    @abstractmethod
    def generate(self) -> str:
        raise NotImplementedError(
            f"{self.__class__.__name__} is a base class and does not implement "
            "'generate' method."
        )



"""Module with HTML generators."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Optional, Union

import soupsavvy.exceptions as exc
from soupsavvy.testing.generators import namespace
from soupsavvy.testing.generators.base import BaseGenerator
from soupsavvy.testing.generators.templates.base import BaseTemplate
from soupsavvy.testing.generators.templates.config import (
    DEFAULT_TEXT_TEMPLATE,
    DEFAULT_VALUE_TEMPLATE,
)
from soupsavvy.testing.generators.templates.templates import ConstantTemplate

# types for generators
TemplateType = Optional[Union[BaseTemplate, str]]

AttributeType = Union["AttributeGenerator", str, tuple[str, TemplateType]]
ChildType = Union["TagGenerator", str]

AttributeList = Iterable[AttributeType]
ChildList = Iterable[ChildType]

AttributeGeneratorInitExceptions = (
    TypeError,
    exc.InvalidTemplateException,
    exc.EmptyNameException,
)


def _get_template_type(
    value: TemplateType, param: str, default: BaseTemplate
) -> BaseTemplate:
    """
    Get the template type from the input value.

    Parameters
    ----------
    value : TemplateType
        The value to get the template type from.
    param : str
        The name of the parameter.
    default : BaseTemplate
        The default template to use if the value is None.

    Returns
    -------
    BaseTemplate
        The template instance.

    Raises
    ------
    InvalidTemplateException
        If the value is not a string, BaseTemplate or None.
    """
    if value is None:
        return default
    elif isinstance(value, str):
        return ConstantTemplate(value)
    elif not isinstance(value, BaseTemplate):
        raise exc.InvalidTemplateException(
            f"{param} must be a string or a BaseTemplate, not {type(value)}"
        )

    return value


class AttributeGenerator(BaseGenerator):
    """
    Class for generating HTML attribute strings.

    AttributeGenerator allows to generate empty HTML attribute with specific name:

    Example
    --------
    >>> gen = AttributeGenerator("class")
    >>> gen.generate()
    'class=""'

    Attribute can also be generated with specified constant value:

    Example
    --------
    >>> gen = AttributeGenerator("class", value="container")
    >>> gen.generate()
    'class="container"'

    Value can be passed as BaseTemplate instance as well:

    Example
    --------
    >>> from soupsavvy.testing.generators import RandomTemplate
    >>> template = RandomTemplate(length=4, seed=42)
    >>> gen = AttributeGenerator("id", value=template)
    >>> gen.generate()
    'id="Nbrn"'

    For more information on available Templates, how to use them
    and customize for your needs, see the documentation.

    See also
    --------
    soupsavvy.testing.generators.templates module
    soupsavvy.testing.generators.TagGenerator class
    """

    def __init__(self, name: str, value: TemplateType = None) -> None:
        """
        Initializes the AttributeGenerator.

        Parameters
        ----------
        name : str
            The name of the attribute.
        value : TemplateType, optional
            The value of the attribute. Defaults to None.
        """
        self._check_name(name)
        self.name = name
        self.value = _get_template_type(
            value,
            param="name",
            default=DEFAULT_VALUE_TEMPLATE,
        )

    def _check_name(self, name: str) -> None:
        """
        Check if the input name is a valid HTML attribute name.

        Parameters
        ----------
        name : str
            The name to check.

        Raises
        ------
        TypeError
            If the name is not a string.
        EmptyNameException
            If the name is an empty string.
        """
        if not isinstance(name, str):
            raise TypeError(f"'name' parameter must be a string, not {type(name)}")

        if not name:
            raise exc.EmptyNameException(
                "Empty string is not allowed html attribute name, "
                "not-empty string must be provided for 'name' parameter."
            )

    def generate(self) -> str:
        """
        Generates the HTML attribute string.

        Returns
        -------
        str
            The generated HTML attribute string.
        """
        value = self.value.generate()
        return f'{self.name}="{value}"'


class TagGenerator(BaseGenerator):
    """
    Class for generating HTML tag strings.

    TagGenerator allows to generate empty HTML tag with specific name:

    Example
    --------
    >>> gen = TagGenerator("div")
    >>> gen.generate()
    '<div></div>'

    Tag can also be generated with specified attributes:

    Example
    --------
    >>> gen = TagGenerator("div", attrs=[("class", "container")])
    >>> gen.generate()
    '<div class="container"></div>'

    Attributes can be passed as an iterable of of mixed types:
    * AttributeGenerator - instance of AttributeGenerator class
    * str - attribute name, value would be a default Template (empty string)
    * tuple[str, TemplateType] - attribute name and literal value or Template

    Example
    --------
    >>> gen = TagGenerator(
    >>>     name="a",
    >>>     attrs=[
    >>>         ("id", "link"),
    >>>         AttributeGenerator("href", "/endpoint"),
    >>>         "class",
    >>>     ],
    >>> )
    >>> gen.generate()
    '<div id="link" href="/endpoint", class=""></div>'

    Similarly, children can be passed as an iterable of mixed types:
    * TagGenerator - instance of TagGenerator class
    * str - tag name, children tags would be empty

    Example
    --------
    >>> gen = TagGenerator(
    >>>     name="div",
    >>>     children=[
    >>>         "a",
    >>>         TagGenerator("span", attrs=[("class", "container")],
    >>>     ],
    >>> )
    >>> gen.generate()
    '<div><a></a><span class="container"></span></div>'

    Text of the tag can be passed as a string or a Template:

    Example
    --------
    >>> gen = TagGenerator("p", text="Hello, World!")
    >>> gen.generate()
    '<p>Hello, World!</p>'

    Example
    --------
    >>> from soupsavvy.testing.generators import RandomTemplate
    >>> template = RandomTemplate(length=4, seed=42)
    >>> gen = TagGenerator("p", text=template)
    >>> gen.generate()
    '<p>Nbrn</p>'

    For more information on available Templates, how to use them
    and customize for your needs, see the documentation.

    Void tags like <img>, <br>, <hr> etc. can be generated as well
    and are automatically closed:

    Example
    --------
    >>> gen = TagGenerator("img", attrs=[("src", "/path/to/image.jpg")])
    >>> gen.generate()
    '<img src="/path/to/image.jpg"/>'

    No children are allowed for void tags, and an error will be raised.

    See also
    --------
    soupsavvy.testing.generators.templates module
    soupsavvy.testing.generators.AttributeGenerator class
    """

    def __init__(
        self,
        name: str,
        attrs: AttributeList = (),
        children: ChildList = (),
        text: TemplateType = None,
    ) -> None:
        """
        Initialize the TagGenerator.

        Parameters
        ----------
        name : str
            The name of the HTML tag.
        attrs : AttributeList, optional
            The attributes of the tag. Defaults to empty tuple.
        children : ChildList, optional
            The children of the tag. Defaults to empty tuple.
        text : TemplateType, optional
            The text content of the tag.
            Defaults to None, which generates empty string.
        """
        if isinstance(attrs, str):
            raise TypeError("'attrs' must be an iterable of attributes, not a string")

        self._void = name in namespace.VOID_TAGS

        if self._void and children:
            raise exc.VoidTagWithChildrenException(
                f"'{name}' is a void tag and cannot have children"
            )

        self._check_name(name)
        self.name = name

        self.text = _get_template_type(
            text,
            param="text",
            default=DEFAULT_TEXT_TEMPLATE,
        )

        self.attributes = self._process_attributes(attrs)
        self.children = [
            child if isinstance(child, TagGenerator) else TagGenerator(child)
            for child in children
        ]

    def _check_name(self, name: str) -> None:
        """
        Check if the input name is a valid HTML tag name.

        Parameters
        ----------
        name : str
            The name to check.

        Raises
        ------
        TypeError
            If the name is not a string.
        EmptyNameException
            If the name is an empty string.
        """
        if not isinstance(name, str):
            raise TypeError(f"'name' parameter must be a string, not {type(name)}")

        if not name:
            raise exc.EmptyNameException(
                "Empty string is not allowed html tag name, "
                "not-empty string must be provided for 'name' parameter."
            )

    def _process_attributes(
        self, attributes: AttributeList
    ) -> list[AttributeGenerator]:
        """
        Process the input attributes of the tag.

        Parameters
        ----------
        attributes : AttributeList
            The attributes to process.

        Returns
        -------
        list[AttributeGenerator]
            The processed instances of AttributeGenerators.

        Raises
        ------
        AttributeParsingError
            If the input attributes could not be parsed into AttributeGenerators.
        """
        # dict approach for uniqueness check
        attr_dict: dict[str, AttributeGenerator] = {}

        for attr in attributes:
            if isinstance(attr, str):
                attr = (attr, None)

            if not isinstance(attr, AttributeGenerator):
                try:
                    attr = AttributeGenerator(*attr)
                except AttributeGeneratorInitExceptions as e:
                    raise exc.AttributeParsingError(
                        f"Attribute {attr} could not be parsed into AttributeGenerator "
                        "due to following error:"
                    ) from e

            attr_name = attr.name

            if attr_name in attr_dict:
                raise exc.NotUniqueAttributesException(
                    f"Input attribute {attr_name} is not unique."
                )

            attr_dict[attr_name] = attr

        return list(attr_dict.values())

    def generate(self) -> str:
        """
        Generates the HTML tag string.

        Returns
        -------
        str
            The generated HTML tag string.
        """
        attrs = " ".join(attr.generate() for attr in self.attributes)
        sep = " " if attrs else ""
        children = "".join(child.generate() for child in self.children)
        tag_content = f"{self.name}{sep}{attrs}"

        if self._void:
            return f"<{tag_content}/>"

        text = self.text.generate()
        return f"<{tag_content}>{children}{text}</{self.name}>"



"""Module with constants and defaults for model classes."""

# default recursive value for finding fields of model within the scope
DEFAULT_RECURSIVE = True
# default strict value for finding fields of model within the scope
DEFAULT_STRICT = False

SCOPE = "__scope__"
INHERIT_FIELDS = "__inherit_fields__"

SPECIAL_FIELDS = {SCOPE, INHERIT_FIELDS}

# based model classes that skip initialization checks
BASE_MODELS = {"BaseModel"}



"""
Module for selectors that represent logical operations on multiple selectors.
These selectors are used to create more complex selectors by combining
multiple selectors in a single object.

Classes
-------
AndSelector - intersection of multiple selectors
NotSelector - negation of a selector (~)
XORSelector - exclusive OR of multiple selectors
OrSelector - union of multiple selectors (|)
"""

from collections import Counter
from functools import reduce
from typing import Optional

from bs4 import Tag

from soupsavvy.selectors.base import CompositeSoupSelector, SoupSelector
from soupsavvy.selectors.combinators import SelectorList
from soupsavvy.selectors.tag_utils import TagIterator, TagResultSet, UniqueTag

# alias for SelectorList
OrSelector = SelectorList


class NotSelector(CompositeSoupSelector):
    """
    Class representing selector of elements that do not match provided selectors.

    Example
    -------
    >>> NotSelector(TypeSelector("div"))

    matches all elements that do not have "div" tag name.

    Example
    -------
    >>> <span> class="widget">Hello World</span> ✔️
    >>> <div class="menu">Hello World</div> ❌

    Object can be initialized with multiple selectors as well, in which case
    at least one selector must match for element to be excluded from the result.

    Object can be created as well by using bitwise NOT operator '~'
    on a SoupSelector object.

    Example
    -------
    >>> ~TypeSelector("div")

    Which is equivalent to the first example.

    This is an equivalent of CSS :not() negation pseudo-class,
    that represents elements that do not match a list of selectors

    Example
    -------
    >>> div:not(.widget) { color: red; }
    >>> :not(strong, .important) { color: red; }

    The second example translated to soupsavvy would be:

    Example
    -------
    >>> NotSelector(TypeSelector("strong"), AttributeSelector("class", "important"))
    >>> ~(TypeSelector("strong") | AttributeSelector("class", "important"))

    Notes
    -----
    For more information on :not() pseudo-class see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:not
    """

    def __init__(
        self,
        selector: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes NotSelectors object with provided positional arguments as selectors.

        Parameters
        ----------
        selectors: SoupSelector
            SoupSelector objects to negate match accepted as positional arguments.
            At least one SoupSelector object is required to create NotSelector.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of SoupSelector.
        """
        self._multiple = bool(selectors)
        super().__init__([selector, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        matching = reduce(
            TagResultSet.__or__,
            (
                TagResultSet(step.find_all(tag, recursive=recursive))
                for step in self.selectors
            ),
        )
        all_tags = list(TagIterator(tag, recursive=recursive))
        result = TagResultSet(all_tags) - matching
        return result.fetch(limit)

    def __invert__(self) -> SoupSelector:
        """
        Overrides __invert__ method to cancel out negation by returning
        the tag in case of single selector, or SoupUnionTag in case of multiple.
        """
        from soupsavvy.selectors.combinators import SelectorList

        if not self._multiple:
            return self.selectors[0]

        return SelectorList(*self.selectors)


class AndSelector(CompositeSoupSelector):
    """
    Class representing an intersection of multiple soup selectors.
    Provides elements matching all of the listed selectors.
    This is `soupsavvy` counterpart of CSS compound selector.

    Example
    -------
    >>> AndSelector(
    ...    TypeSelector("div"),
    ...    AttributeSelector(name="class", value="widget")
    ... )

    matches all elements that have "div" tag name AND 'class' attribute "widget".

    Example
    -------
    >>> <div class="widget">Hello World</div> ✔️
    >>> <span class="widget">Hello World</span> ❌
    >>> <div class="menu">Hello World</div> ❌

    Object can be initialized with multiple selectors as well, in which case
    all selectors must match for element to be included in the result.

    Object can be created as well by using bitwise AND operator '&'
    on two SoupSelector objects.

    Example
    -------
    >>> TypeSelector("div") & AttributeSelector(name="class", value="widget")

    Which is equivalent to the first example.

    This is an equivalent of CSS selectors concatenation.

    Example
    -------
    >>> div.class1#id1 { color: red; }

    which translated to soupsavvy would be:

    Example
    -------
    >>> TypeSelector("div") & ClassSelector("class1") & IdSelector("id1")

    Notes
    -----
    For more information on compound selectors see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_selectors/Selector_structure#compound_selector
    """

    def __init__(
        self,
        selector1: SoupSelector,
        selector2: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes AndSelector object with provided
        positional arguments as selectors.

        Parameters
        ----------
        selectors: SoupSelector
            SoupSelector objects to match accepted as positional arguments.
            At least two SoupSelector objects are required to create AndSelector.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of SoupSelector.
        """
        super().__init__([selector1, selector2, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        matching = reduce(
            TagResultSet.__and__,
            (
                TagResultSet(step.find_all(tag, recursive=recursive))
                for step in self.selectors
            ),
        )
        return matching.fetch(limit)


class XORSelector(CompositeSoupSelector):
    """
    Class representing an exclusive OR of multiple soup selectors.
    Element is selected if it matches exactly one of the provided selectors.

    Example
    -------
    >>> XORSelector(TypeSelector("div"), AttributeSelector(class="widget"))

    Matches all elements that have either "div" tag name or 'class' attribute "widget".
    Elements with "div" tag name and 'class' attribute "widget" do not match selector.

    This is a shortcut for defining XOR operation between two selectors like this:

    Example
    -------
    >>> selector1 = TypeSelector("div")
    ... selector2 = AttributeSelector("class", value="widget")
    ... xor = (selector1 & (~selector2)) | ((~selector1) & selector2)

    Object can be initialized with more then two selectors as well, in which case
    exactly one selector must match for element to be included in the result.
    """

    def __init__(
        self,
        selector1: SoupSelector,
        selector2: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes XORSelector object with provided positional arguments as selectors.

        Parameters
        ----------
        selectors: SoupSelector
            SoupSelector objects to match accepted as positional arguments.
            At least two selector are required to create XORSelector.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of SoupSelector.
        """
        super().__init__([selector1, selector2, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        unique = (
            UniqueTag(element)
            for step in self.selectors
            for element in step.find_all(tag, recursive=recursive)
        )
        results = TagResultSet(
            [element.tag for element, count in Counter(unique).items() if count == 1]
        )
        # keep order of tags and limit
        results = TagResultSet(list(TagIterator(tag, recursive=recursive))) & results
        return results.fetch(limit)



"""
Module for css selectors API and not only.
Imitate css some pseudo-classes behavior done on soupsavvy selectors,
and provides other shortcut functions for creating composite selectors.

Contains:

* is_: function to create SelectorList from list of selectors
* where: alias for is_, no difference in behavior in scraping context
* or_: alias for is_, no difference in behavior in scraping context
* not_: function to create NotSelector from list of selectors
* and_: function to create AndSelector from list of selectors
* has: function to create HasSelector from list of selectors
* xor: function to create XORSelector from list of selectors

There is not similar functionality in css for AndSelector expect for selector concatenation,
and_ is more to have and alternative way to create AndSelector for soupsavvy selectors.

These utils do not brings extra functionality into the package, but they can be
used for convenience and readability in some cases.
"""

from soupsavvy.selectors.base import SoupSelector
from soupsavvy.selectors.combinators import SelectorList
from soupsavvy.selectors.logical import AndSelector, NotSelector, XORSelector
from soupsavvy.selectors.relative import HasSelector


def is_(
    selector1: SoupSelector,
    selector2: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> SelectorList:
    """
    Returns an union of multiple soup selectors.
    Provides elements matching any of the selectors in the union.

    Parameters
    ----------
    selectors: SoupSelector
        SoupSelector objects to match accepted as positional arguments.
        At least two objects are required per SelectorList class requirements.

    Returns
    -------
    SelectorList
        SelectorList object that represents union of provided selectors.

    Example
    -------
    >>> is_(TypeSelector("h1"), TypeSelector("h2"))
    >>> where(TypeSelector("h1"), TypeSelector("h2"))

    This is an equivalent of CSS comma selector (selector list) or :is() selector.
    Function is also aliased by 'where' function that imitate css :where pseudo-class.
    In css the main difference is their specificity, but in this context they are the same.

    Example
    -------
    >>> h1, h1 { color: red;}
    >>> :is(h1, h2) { color: red; }
    >>> :where(h1, h2) { color: red; }

    Notes
    -----
    For more information on selector list see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:is
    https://developer.mozilla.org/en-US/docs/Web/CSS/:where
    https://developer.mozilla.org/en-US/docs/Web/CSS/Selector_list
    """
    return SelectorList(selector1, selector2, *selectors)


where = is_
or_ = is_


def not_(selector: SoupSelector, /, *selectors: SoupSelector) -> NotSelector:
    """
    Returns an negation of multiple soup selectors.
    NotSelector represents elements that do not match a list of selectors.

    Parameters
    ----------
    selectors: SoupSelector
        SoupSelector objects to negate match accepted as positional arguments.

    Returns
    -------
    NotSelector
        NotSelector object that represents negation of provided selectors.

    Example
    -------
    >>> not_(TypeSelector("div"), TypeSelector("a"))

    This is an equivalent of CSS :not() negation pseudo-class,
    that represents elements that do not match a list of selectors

    Example
    -------
    >>> :not(div, a) { color: red; }

    Notes
    -----
    For more information on selector list see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:not
    """
    return NotSelector(selector, *selectors)


def and_(
    selector1: SoupSelector,
    selector2: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> AndSelector:
    """
    Returns an intersection of multiple soup selectors.
    Element must match all of them to be selected.

    Parameters
    ----------
    selectors: SoupSelector
        SoupSelector objects to match accepted as positional arguments.
        At least two objects are required per AndSelector class requirements.

    Returns
    -------
    AndSelector
        AndSelector object that represents intersection of provided selectors.

    Example
    -------
    >>> and_(TypeSelector("div"), AttributeSelector(name="class", value="widget"))

    This is an equivalent of CSS selectors concatenation.

    Example
    -------
    >>> div.class1#id1 { color: red; }
    """
    return AndSelector(selector1, selector2, *selectors)


def has(
    selector: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> HasSelector:
    """
    Returns HasSelector whose behavior imitate css :has() pseudo-class,
    that represents elements if any of the relative selectors that are passed as an argument
    match at least one element when anchored against this element.

    Parameters
    ----------
    selectors: SoupSelector
        SoupSelector objects to match accepted as positional arguments.
        At least two objects are required per AndSelector class requirements.

    Returns
    -------
    HasSelector
        HasSelector object that represents :has() pseudo-class behavior.

    Example
    -------
    >>> has_(TypeSelector("div"), TypeSelector("div"))

    This is an equivalent of CSS :has() pseudo-class:

    Example
    -------
    >>> :has(div, a) { color: red; }

    For now only default combinator (descendant) for 'relative' selectors is supported,
    so imitating css behavior of this kind is not possible yet.

    Example
    -------
    >>> :has(+ div, > a) { color: red; }

    Notes
    -----
    For more information on :has() pseudo-class see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:has
    """
    return HasSelector(selector, *selectors)


def xor(
    selector1: SoupSelector,
    selector2: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> XORSelector:
    """
    Returns an symmetric difference of multiple soup selectors.
    Element must match exactly one of them to be selected.

    Parameters
    ----------
    selectors: SoupSelector
        SoupSelector objects to match accepted as positional arguments.
        At least two objects are required per XORSelector class requirements.

    Returns
    -------
    XORSelector
        XORSelector object that represents symmetric difference of provided selectors.

    Example
    -------
    >>> xor(TypeSelector("div"), AttributeSelector(name="class", value="widget"))
    """
    return XORSelector(selector1, selector2, *selectors)



from .attributes import AttributeSelector, ClassSelector, IdSelector
from .combinators import (
    AncestorCombinator,
    ChildCombinator,
    DescendantCombinator,
    NextSiblingCombinator,
    ParentCombinator,
    SelectorList,
    SubsequentSiblingCombinator,
)
from .general import (
    AnyTagSelector,
    PatternSelector,
    SelfSelector,
    TypeSelector,
    UniversalSelector,
)
from .logical import AndSelector, NotSelector, OrSelector, XORSelector
from .relative import Anchor, HasSelector

__all__ = [
    "UniversalSelector",
    "AttributeSelector",
    "TypeSelector",
    "PatternSelector",
    "SelfSelector",
    "SelectorList",
    "DescendantCombinator",
    "AncestorCombinator",
    "ParentCombinator",
    "AndSelector",
    "NotSelector",
    "ChildCombinator",
    "NextSiblingCombinator",
    "SubsequentSiblingCombinator",
    "OrSelector",
    "HasSelector",
    "Anchor",
    "IdSelector",
    "ClassSelector",
    "XORSelector",
    "AnyTagSelector",
]



"""Module for base classes and interfaces for operations."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Optional

from bs4 import Tag

import soupsavvy.exceptions as exc
from soupsavvy.interfaces import Comparable, Executable, TagSearcher

if TYPE_CHECKING:
    from soupsavvy.operations.general import OperationPipeline


def check_operation(x: Any, message: Optional[str] = None) -> BaseOperation:
    """
    Checks if provided object is a valid `soupsavvy` operation.
    Checks for instance of BaseOperation and raises an exception if not.
    Returns provided object if fulfills the condition for convenience.

    Parameters
    ----------
    x : Any
        Any object to be validated as correct operation.
    message : str, optional
        Custom message to be displayed in case of raising an exception.
        By default None, which results in default message.

    Raises
    ------
    NotOperationException
        If provided object is not an instance of BaseOperation.
    """
    message = message or f"Object {x} is not an instance of {BaseOperation.__name__}."

    if not isinstance(x, BaseOperation):
        raise exc.NotOperationException(message)

    return x


class BaseOperation(Executable, Comparable):
    """
    Base class for all `soupsavvy` operations.
    Operations are used to process the selection results from the soup,
    extract and transform the data.

    Operations can be chained together using the pipe operator '|'.

    Example
    -------
    >>> from soupsavvy.operations.general import Operation
    ... operation = Operation(str.lower) | Operation(str.strip)
    ... operation.execute("  TEXT  ")
    'text'

    Operations can be combined with selectors to extract and transform
    target information.

    Example
    -------
    >>> from soupsavvy import TypeSelector
    ... from soupsavvy.operations import Operation, Text
    ... selector = TypeSelector("div") | Text() | Operation(int)
    ... selector.find(soup)
    42

    BaseOperation inherits from `Comparable` interface, `__eq__` method needs to be
    implemented in derived classes.
    """

    def execute(self, arg: Any) -> Any:
        """
        Execute the operation on the given argument and return the result.

        Parameters
        ----------
        arg : Any
            Argument to be processed by the operation.

        Returns
        -------
        Any
            Result of the operation.
        """
        try:
            return self._execute(arg)
        except Exception as e:
            raise exc.FailedOperationExecution(
                f"Failed to execute operation: {e}"
            ) from e

    @abstractmethod
    def _execute(self, arg: Any) -> Any:
        """Internal method for executing the operation."""

        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface, "
            "and does not implement this method."
        )

    def __or__(self, x: Any) -> OperationPipeline:
        """
        Overrides __or__ method called also by pipe operator '|'.
        Syntactical Sugar for logical disjunction, that creates OperationPipeline,
        which chains two operations together.

        Parameters
        ----------
        x : BaseOperation
            BaseOperation object to be combined into new OperationPipeline.

        Returns
        -------
        OperationPipeline
            New `OperationPipeline` with extended operations.

        Raises
        ------
        NotOperationException
            If provided object is not of type BaseOperation.
        """
        from soupsavvy.operations.general import OperationPipeline

        message = (
            f"Bitwise OR not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {BaseOperation.__name__}."
        )
        other = check_operation(x, message=message)
        return OperationPipeline(self, other)


class OperationSearcherMixin(BaseOperation, TagSearcher):
    """
    Mixin of `BaseOperation` and `TagSearcher` interfaces.
    Allows operations to be used as field searchers in model
    to perform operation directly on scope element.
    """

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Any]:
        """
        Extracts information from provided element.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object to extract text from.
        recursive : bool, optional
            Ignored, for consistency with interface.
        limit : int, optional
            Ignored, for consistency with interface.

        Returns
        -------
        list[Any]
            Extracted information from the tag.
        """
        return [self.execute(tag)]

    def find(
        self,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Any:
        """
        Extracts information from provided element.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object to extract text from.
        strict : bool, optional
            Ignored, for consistency with interface.
        recursive : int, optional
            Ignored, for consistency with interface.

        Returns
        -------
        Any
            Extracted information from the tag.
        """
        return self.execute(tag)



"""Module for base classes and interfaces for tag selectors."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal, Optional, Union, overload

from bs4 import NavigableString, Tag
from typing_extensions import deprecated

import soupsavvy.exceptions as exc
import soupsavvy.selectors.namespace as ns
from soupsavvy.interfaces import Comparable, TagSearcher

if TYPE_CHECKING:
    from soupsavvy.operations.base import BaseOperation
    from soupsavvy.operations.selection_pipeline import SelectionPipeline
    from soupsavvy.selectors.combinators import (
        AncestorCombinator,
        ChildCombinator,
        DescendantCombinator,
        NextSiblingCombinator,
        ParentCombinator,
        SelectorList,
        SubsequentSiblingCombinator,
    )
    from soupsavvy.selectors.logical import AndSelector


def check_selector(x: Any, message: Optional[str] = None) -> SoupSelector:
    """
    Checks if provided object is a valid `soupsavvy` selector.
    Checks for instance of SoupSelector and raises an exception if not.
    Returns provided object if fulfills the condition for convenience.

    Parameters
    ----------
    x : Any
        Any object to be validated as correct selector.
    message : str, optional
        Custom message to be displayed in case of raising an exception.
        By default None, which results in default message.

    Raises
    ------
    NotSoupSelectorException
        If provided object is not an instance of SoupSelector.
    """
    message = message or f"Object {x} is not an instance of {SoupSelector.__name__}."

    if not isinstance(x, SoupSelector):
        raise exc.NotSoupSelectorException(message)

    return x


class SoupSelector(TagSearcher, Comparable):
    """
    Interface for classes that define tags and provide functionality
    to find them in BeautifulSoup Tags.

    Methods
    -------
    find(tag: Tag, strict: bool = False)
        Finds tag element in BeautifulSoup Tag by wrapping Tag find method
        and passing parameters specific to given tag.
        If strict parameter is set to True and element is not found exception is raised.
        If strict is set to False, which is a default and element is not found
        result of Tag.find is return, which is None.
        Otherwise matching Tag is returned.
    find_all(tag: Tag)
        Finds all tag elements in BeautifulSoup Tag by wrapping Tag find_all method
        and passing parameters specific to given tag.

    Notes
    -----
    To implement SoupSelector interface child class must implement:
    * 'find_all' method that returns a list of matching elements in bs4.Tag.
    It could optionally implement:
    * '_find' method that returns result of bs4.Tag 'find' method.
    By default '_find' method is implemented by calling 'find_all' method
    and returning first element if any found or None otherwise.
    If different logic is required or performance can be improved, '_find' method
    can be overridden in child class.
    """

    @overload
    def find(
        self,
        tag: Tag,
        strict: Literal[False] = ...,
        recursive: bool = ...,
    ) -> Optional[Tag]: ...

    @overload
    def find(
        self,
        tag: Tag,
        strict: Literal[True] = ...,
        recursive: bool = ...,
    ) -> Tag: ...

    @overload
    def find(
        self,
        tag: Tag,
        strict: bool = ...,
        recursive: bool = ...,
    ) -> Optional[Tag]: ...

    def find(
        self,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Optional[Tag]:
        """
        Finds a first matching element in provided BeautifulSoup Tag.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object.
        strict : bool, optional
            If True, raises an exception if tag was not found in markup,
            if False and tag was not found, returns None by defaulting to BeautifulSoup
            implementation. Value of this parameter does not affect behavior if tag
            was successfully found in the markup. By default False.
        recursive : bool, optional
            bs4.Tag.find method parameter that specifies if search should be recursive.
            If set to False, only direct children of the tag will be searched.
            By default True.

        Returns
        -------
        Tag | None
            BeautifulSoup Tag if it tag was found in the markup, else None.

        Notes
        -----
        If result is an instance of NavigableString,
        exception is raised to avoid any type of inconsistency downstream.

        Raises
        ------
        TagNotFoundException
            If strict parameter is set to True and tag was not found in markup.
        NavigableStringException
            If NavigableString was returned by bs4.
        """
        result = self._find(tag, recursive=recursive)

        if result is None:
            if strict:
                raise exc.TagNotFoundException("Tag was not found in markup.")
            return None

        if isinstance(result, NavigableString):
            raise exc.NavigableStringException(
                f"NavigableString was returned for {result} string search, "
                "invalid operation for SoupSelector find."
            )

        return result

    @abstractmethod
    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        """
        Finds all matching elements in provided BeautifulSoup Tag.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object.
        recursive : bool, optional
            bs4.Tag.find method parameter that specifies if search should be recursive.
            If set to False, only direct children of the tag will be searched.
            By default True.
        limit : int, optional
            bs4.Tag.find_all method parameter that specifies maximum number of elements
            to return. If set to None, all elements are returned. By default None.

        Returns
        -------
        list[Tag]
            A list of Tag objects matching tag specifications.
            If none found, the list is empty.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface, "
            "and does not implement this method."
        )

    def _find(self, tag: Tag, recursive: bool = True) -> ns.FindResult:
        """
        Returns an object that is a result of markup search.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object.
        recursive : bool, optional
            bs4.Tag.find method parameter that specifies if search should be recursive.
            If set to False, only direct children of the tag will be searched.
            By default True.

        Returns
        -------
        NavigableString | Tag | None:
            Result of bs4.Tag find method with tag parameters.
        """
        elements = self.find_all(tag, recursive=recursive, limit=1)
        return elements[0] if elements else None

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """
        Check self and other object for equality.

        This method is abstract and must be implemented by all selectors.
        Selectors are considered equal if their find methods return the same result.

        Calling find or find_all methods on selectors that are equal
        should return the same results.

        Example
        -------
        >>> selector1 = TypeSelector("div")
        >>> selector2 = TypeSelector("div")
        >>> selector1 == selector2
        True
        >>> selector1.find(tag) == selector2.find(tag)
        True
        """

        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface, "
            "and does not implement this method."
        )

    @overload
    def __or__(self, x: SoupSelector) -> SelectorList: ...

    @overload
    def __or__(self, x: BaseOperation) -> SelectionPipeline: ...

    def __or__(
        self, x: Union[SoupSelector, BaseOperation]
    ) -> Union[SelectorList, SelectionPipeline]:
        """
        Overrides __or__ method called also by pipe operator '|'.
        Syntactical Sugar for logical disjunction, that creates SelectorList,
        which matches at least one of the elements.

        Parameters
        ----------
        x : SoupSelector
            SoupSelector object to be combined into new SelectorList with current one.

        Returns
        -------
        SelectorList
            Union of two SoupSelector that can be used to get all elements
            that match at least one of the elements in the Union.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not of SoupSelector type.
        """
        from soupsavvy.operations.base import BaseOperation
        from soupsavvy.operations.selection_pipeline import SelectionPipeline
        from soupsavvy.selectors.combinators import SelectorList

        if isinstance(x, BaseOperation):
            return SelectionPipeline(selector=self, operation=x)

        message = (
            f"Bitwise OR not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__} or {BaseOperation.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, SelectorList):
            args = [*self.selectors, x]
            # return new SelectorList with updated steps
            return SelectorList(*args)

        return SelectorList(self, x)

    def __invert__(self) -> SoupSelector:
        """
        Overrides __invert__ method called also by tilde operator '~'.
        Syntactical Sugar for bitwise NOT operator, that creates a NotSelector
        matching everything that is not matched by the SoupSelector.

        Parameters
        ----------
        x : SoupSelector
            SoupSelector object to be negated into new NotSelector.

        Returns
        -------
        NotSelector
            Negation of SoupSelector that can be used to get all elements
            that do not match the SoupSelector.
        SoupSelector
            Any SoupSelector object in case of directly inverting
            NotSelector to get original SoupSelector.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not of SoupSelector type.
        """
        from soupsavvy.selectors.logical import NotSelector

        return NotSelector(self)

    def __and__(self, x: SoupSelector) -> AndSelector:
        """
        Overrides __and__ method called also by ampersand operator '&'.
        Syntactical Sugar for bitwise AND operator, that creates an AndSelector
        matching everything that is matched by both SoupSelector.

        Parameters
        ----------
        x : SoupSelector
            SoupSelector object to be combined into new AndSelector
            with current SoupSelector.

        Returns
        -------
        AndSelector
            Intersection of two SoupSelector that can be used to get all elements
            that match both SoupSelector.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not of SoupSelector type.
        """
        from soupsavvy.selectors.logical import AndSelector

        message = (
            f"Bitwise AND not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, AndSelector):
            args = [*self.selectors, x]
            # return new AndSelector with updated steps
            return AndSelector(*args)

        return AndSelector(self, x)

    def __gt__(self, x: SoupSelector) -> ChildCombinator:
        """
        Overrides __gt__ method called also by greater-than operator '>'.
        Syntactical Sugar for greater-than operator, that creates a ChildCombinator
        which is equivalent to child combinator in css.

        Example
        -------
        >>> div > a

        matches all 'a' elements that are direct children of 'div' elements.
        The same can be achieved by using '>' operator on two SoupSelector objects.

        Example
        -------
        >>> TypeSelector("div") > TypeSelector("a")

        Which results in

        Example
        -------
        >>> ChildCombinator(TypeSelector("div"), TypeSelector("a"))

        Parameters
        ----------
        x : SoupSelector
            SoupSelector object to be combined into new ChildCombinator
            as selector for children of elements matched by current SoupSelector.

        Returns
        -------
        ChildCombinator
            ChildCombinator object that can be used to get all matching elements
            that are direct children of the elements matched by current SoupSelector.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not of SoupSelector type.
        """
        from soupsavvy.selectors.combinators import ChildCombinator

        message = (
            f"GT operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, ChildCombinator):
            args = [*self.selectors, x]
            # return new ChildCombinator with updated steps
            return ChildCombinator(*args)

        return ChildCombinator(self, x)

    def __lt__(self, x: SoupSelector) -> ParentCombinator:
        """
        Overrides __lt__ method called also by less-than operator '<'.
        Syntactical Sugar for less-than operator, that creates a ParentCombinator.

        Example
        -------
        >>> TypeSelector("a") < TypeSelector("div")

        Which results in

        Example
        -------
        >>> ParentCombinator(TypeSelector("a"), TypeSelector("div"))

        Matches all 'div' elements that are parents of 'a' elements.

        Parameters
        ----------
        x : SoupSelector
            SoupSelector object to be combined into new ParentCombinator
            as selector for parent of elements matched by current SoupSelector.

        Returns
        -------
        ParentCombinator
            ParentCombinator object that can be used to get all matching elements
            that are parents of the elements matched by current SoupSelector.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not of SoupSelector type.
        """
        from soupsavvy.selectors.combinators import ParentCombinator

        message = (
            f"LT operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, ParentCombinator):
            args = [*self.selectors, x]
            # return new ParentCombinator with updated steps
            return ParentCombinator(*args)

        return ParentCombinator(self, x)

    def __add__(self, x: SoupSelector) -> NextSiblingCombinator:
        """
        Overrides __add__ method called also by plus operator '+'.
        Syntactical Sugar for addition operator, that creates a NextSiblingCombinator
        which is equivalent to next sibling element css selector.

        Example
        -------
        >>> div + a

        matches all 'a' elements that immediately follow 'div' elements,
        it means that both elements are children of the same parent element.

        The same can be achieved by using '+' operator on two SoupSelector objects.

        Example
        -------
        >>> TypeSelector("div") + TypeSelector("a")

        Which results in

        Example
        -------
        >>> NextSiblingCombinator(TypeSelector("div"), TypeSelector("a"))

        Parameters
        ----------
        x : SoupSelector
            SoupSelector object to be combined into new NextSiblingCombinator
            as selector for next sibling of current SoupSelector.

        Returns
        -------
        NextSiblingCombinator
            NextSiblingCombinator object that can be used to get all matching elements
            that are next siblings of the current SoupSelector.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not of SoupSelector type.
        """
        from soupsavvy.selectors.combinators import NextSiblingCombinator

        message = (
            f"ADD operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, NextSiblingCombinator):
            args = [*self.selectors, x]
            # return new NextSiblingCombinator with updated steps
            return NextSiblingCombinator(*args)

        return NextSiblingCombinator(self, x)

    def __mul__(self, x: SoupSelector) -> SubsequentSiblingCombinator:
        """
        Overrides __mul__ method called also by multiplication operator '*'.
        Syntactical Sugar for multiplication operator, that creates
        a SubsequentSiblingCombinator which is equivalent to subsequent sibling
        element css selector.

        Example
        -------
        >>> div ~ a

        matches all 'a' elements that follow 'div' elements and share the same parent.
        The same can be achieved by using '*' operator on two SoupSelector objects.

        Example
        -------
        >>> TypeSelector("div") * TypeSelector("a")

        Which results in

        Example
        -------
        >>> SubsequentSiblingCombinator(TypeSelector("div"), TypeSelector("a"))

        Parameters
        ----------
        x : SoupSelector
            SoupSelector object to be combined into new SubsequentSiblingCombinator
            as selector for following sibling of current SoupSelector.

        Returns
        -------
        SubsequentSiblingCombinator
            SubsequentSiblingCombinator object that can be used to get all
            matching elements that are following siblings of the current SoupSelector.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not of SoupSelector type.
        """
        from soupsavvy.selectors.combinators import SubsequentSiblingCombinator

        message = (
            f"MUL operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, SubsequentSiblingCombinator):
            args = [*self.selectors, x]
            # return new SubsequentSiblingCombinator with updated steps
            return SubsequentSiblingCombinator(*args)

        return SubsequentSiblingCombinator(self, x)

    def __rshift__(self, x: SoupSelector) -> DescendantCombinator:
        """
        Overrides __rshift__ method called also by right shift operator '>>'.
        Syntactical Sugar for right shift operator, that creates a DescendantCombinator
        which is equivalent to descendant combinator in css, typically represented
        by a single space character.

        Example
        -------
        >>> div a

        matches all 'a' elements with 'div' as their ancestor.

        Example
        -------
        >>> TypeSelector("div") >> TypeSelector("a")

        Which results in

        Example
        -------
        >>> DescendantCombinator(TypeSelector("div"), TypeSelector("a"))

        Parameters
        ----------
        x : SoupSelector
            SoupSelector object to be combined into new DescendantCombinator
            as selector for descendant of current SoupSelector.

        Returns
        -------
        DescendantCombinator
            DescendantCombinator object that can be used to get all matching elements
            that are descendants of the current SoupSelector.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not of SoupSelector type.

        Notes
        -----
        For more information on descendant combinator see:
        https://developer.mozilla.org/en-US/docs/Web/CSS/Descendant_combinator
        """
        from soupsavvy.selectors.combinators import DescendantCombinator

        message = (
            f"RIGHT SHIFT operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, DescendantCombinator):
            args = [*self.selectors, x]
            # return new DescendantCombinator with updated steps
            return DescendantCombinator(*args)

        return DescendantCombinator(self, x)

    def __lshift__(self, x: SoupSelector) -> AncestorCombinator:
        """
        Overrides __lshift__ method called also by left shift operator '<<'.
        Syntactical Sugar for left-shift operator, that creates an AncestorCombinator.

        Example
        -------
        >>> TypeSelector("a") << TypeSelector("div")

        Which results in

        Example
        -------
        >>> AncestorCombinator(TypeSelector("a"), TypeSelector("div"))

        Matches all 'div' elements that are ancestors of 'a' elements.

        Parameters
        ----------
        x : SoupSelector
            SoupSelector object to be combined into new AncestorCombinator
            as ancestor selector of current SoupSelector.

        Returns
        -------
        AncestorCombinator
            AncestorCombinator object that can be used to get all matching elements
            that are ancestors of the elements matched by current SoupSelector.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not of SoupSelector type.
        """
        from soupsavvy.selectors.combinators import AncestorCombinator

        message = (
            f"LEFT SHIFT operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, AncestorCombinator):
            args = [*self.selectors, x]
            # return new AncestorCombinator with updated steps
            return AncestorCombinator(*args)

        return AncestorCombinator(self, x)


class SelectableCSS(ABC):
    """
    Interface for Tags that can be searched with css selector.

    Notes
    -----
    To implement SelectableCSS interface, child class must implement:
    * 'selector' property, which return a string representing element css selector.
    """

    @property
    @abstractmethod
    def css(self) -> str:
        """Returns string representing element css selector."""
        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface, "
            "and does not implement this property."
        )

    @property
    @deprecated("'selector' property is deprecated, use 'css' instead.")
    def selector(self) -> str:
        """Returns string representing element css selector."""
        return self.css


class CompositeSoupSelector(SoupSelector):
    """
    Interface for Tags that uses multiple steps to find elements.

    Notes
    -----
    To implement CompositeSoupSelector interface, child class must implement
    call super init method with provided tags as arguments to check and assign them.

    Attributes
    ----------
    selectors : list[SoupSelector]
        List of SoupSelector objects passed to CompositeSoupSelector.
    """

    # if the order of selectors is not relevant - by default True
    COMMUTATIVE = True

    def __init__(self, selectors: Iterable[SoupSelector]) -> None:
        """
        Initializes CompositeSoupSelector object with provided tags.
        Checks if all tags are instances of SoupSelector and assigns
        them to 'selectors' attribute.

        Parameters
        ----------
        selectors: Iterable[SoupSelector]
            SoupSelector objects passed to CompositeSoupSelector.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of SoupSelector.
        """
        self._selectors = [check_selector(selector) for selector in selectors]

    @property
    def selectors(self) -> list[SoupSelector]:
        """
        Returns a list of selectors that composite selector consists of.

        Returns
        -------
        list[SoupSelector]
            List of SoupSelector objects.
        """
        return self._selectors

    def __eq__(self, other: object) -> bool:
        # check for CompositeSoupSelector type for type checking sake
        if not isinstance(other, CompositeSoupSelector):
            return False
        elif type(self) is not type(other):
            # checking for exact type match - isinstance(other, self.__class__)
            # when other is subclass of self.__class__ would call other.__eq__(self)
            # which is not desired behavior, as it returns False
            return False

        if not self.__class__.COMMUTATIVE:
            return self.selectors == other.selectors

        # if order is irrelevant, check if all selectors are covered by other
        # and vice versa - selector lists can be of different size as well
        return all(
            any(self_selector == other_selector for other_selector in other.selectors)
            for self_selector in self.selectors
        ) and all(
            any(other_selector == self_selector for self_selector in self.selectors)
            for other_selector in other.selectors
        )

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(map(str, self.selectors))})"

    def __repr__(self) -> str:
        return str(self)



"""
Module for general purpose selectors.

Classes
-------
TypeSelector - combines type and attribute selectors
PatternSelector - matches elements based on text content and selector
UniversalSelector - universal selector (*)
"""

import itertools
from dataclasses import dataclass
from typing import Optional, Pattern

from bs4 import SoupStrainer, Tag
from typing_extensions import deprecated

import soupsavvy.selectors.namespace as ns
from soupsavvy.selectors.base import SelectableCSS, SoupSelector
from soupsavvy.selectors.tag_utils import TagIterator


@dataclass
class TypeSelector(SoupSelector, SelectableCSS):
    """
    Component for finding elements based on tag name.
    `soupsavvy` counterpart of css type selector.

    Example
    -------
    >>> TypeSelector("div")

    matches all elements that have "div" tag name.

    Example
    -------
    >>> <div class="widget">Hello World</div> ✔️
    >>> <a href="/shop">Hello World</a> ❌

    It's soupsavvy counterpart of passing name to bs4 find methods.

    Example
    -------
    >>> soup.find_all("div")

    Returns the same result as:

    Example
    -------
    >>> TypeSelector("div").find_all(soup)

    Parameters
    ----------
    name : str
        HTML tag name ex. "a", "div".

    Notes
    -----
    For more information about type selectors see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/Type_selectors
    """

    name: str

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        params = {ns.NAME: self.name}
        return tag.find_all(**params, recursive=recursive, limit=limit)

    @property
    def css(self) -> str:
        # css selector for tag name is just the tag name ex. "div"
        return self.name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TypeSelector):
            return False

        # TypeSelectors produce the same results if names of the tag are the same
        return self.name == other.name


@dataclass
class PatternSelector(SoupSelector):
    """
    Class representing HTML element with specific string pattern for text.
    Provides elements matching with text matching the pattern.

    Example
    -------
    >>> PatternSelector(pattern="Hello World")

    matches all element with exact text content "Hello World".

    Example
    -------
    >>> <div>Hello World</div> ✔️
    >>> <div>Hello Python</div> ❌
    >>> <div>Hello World 3</div> ❌

    Parameters
    ----------
    pattern: str | Pattern
        A pattern to match text of the element. Can be a string for exact match
        or Pattern for any more complex regular expressions.

    Notes
    -----
    Selector uses re.search function to match text content
    if compiled regex is passed as pattern.
    Providing 're.compile(r"[0-9]")' as pattern will much any element with a digit in text.

    Example
    -------
    >>> import re
    >>> PatternSelector(pattern=re.compile(r"[0-9]"))

    Example
    -------
    >>> <div>Hello World 123</div> ✔️
    >>> <div>Hello World</div> ❌

    Due to bs4 implementation, element does not match the pattern if it has any children.
    Only leaf nodes can be returned by PatternSelector find methods.

    Example
    -------
    >>> <div>Hello World<span></span></div> ❌
    """

    pattern: ns.PatternType

    def __post_init__(self) -> None:
        """Sets up compiled regex pattern used for SoupStrainer in find methods."""
        self._pattern = (
            str(self.pattern) if not isinstance(self.pattern, Pattern) else self.pattern
        )

    def find_all(
        self, tag: Tag, recursive: bool = True, limit: Optional[int] = None
    ) -> list[Tag]:
        iterator = TagIterator(tag, recursive=recursive)
        strainer = SoupStrainer(string=self._pattern)  # type: ignore
        filter_ = filter(strainer.search_tag, iterator)
        return list(itertools.islice(filter_, limit))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PatternSelector):
            return False

        return self._pattern == other._pattern


@dataclass
class UniversalSelector(SoupSelector, SelectableCSS):
    """
    Class representing a wildcard tag that matches ny and all types of elements
    in html page.

    UniversalSelector implements SelectableCSS interface with wildcard css selector "*",
    aka. universal selector, that matches all elements in html page.

    Example
    -------
    >>> any_element = UniversalSelector()
    >>> any_element.selector
    "*"

    Notes
    -----
    For more information on universal selector see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/Universal_selectors
    """

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        return tag.find_all(recursive=recursive, limit=limit)

    @property
    def css(self) -> str:
        """Returns wildcard css selector matching all elements in the markup."""
        return ns.CSS_SELECTOR_WILDCARD

    def __eq__(self, other: object) -> bool:
        # UniversalSelector is always the same
        return isinstance(other, UniversalSelector)


@deprecated(f"AnyTagSelector is deprecated, use 'UniversalSelector' class instead.")
class AnyTagSelector(UniversalSelector):
    """Alias for UniversalSelector class. Deprecated component."""


class SelfSelector(SoupSelector):
    """
    Class representing a selector that matches the element itself.
    Convenience component that can be used for compatibility.

    Example
    -------
    >>> SelfSelector()

    always matches the tag that is passed to the find methods.

    Can be used in user-defined model for scope if element itself is the scope.
    """

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        return [tag]

    def __eq__(self, other: object) -> bool:
        # needs to be the same type
        return isinstance(other, SelfSelector)



"""
Module for selectors that are based on HTML element attributes.

They are used for selection of various elements in the markup,
based on attributes values.

Classes
-------
AttributeSelector - Selects element for any attribute value.
IdSelector - Selects element based on 'id' attribute value.
ClassSelector - Selects element based on 'class' attribute value.
"""

import re
from dataclasses import dataclass
from typing import Optional, Pattern

from bs4 import Tag

import soupsavvy.selectors.namespace as ns
from soupsavvy.selectors.base import SoupSelector
from soupsavvy.selectors.namespace import PatternType


@dataclass
class AttributeSelector(SoupSelector):
    """
    Class representing attribute of the HTML element.
    `soupsavvy` counterpart of css attribute selectors, where any substring value
    syntax is handled by regex pattern matching.

    Example
    -------
    >>> AttributeSelector(name="class", value="widget")

    matches all elements that have 'class' attribute with value "widget".

    Example
    -------
    >>> <div class="widget">Hello World</div> ✔️
    >>> <div class="menu">Hello World</div> ❌

    Parameters
    ----------
    name : str
        HTML element attribute name ex. "class", "href"
    value : str | Pattern, optional
        Value of the attribute, like a class name. Can be regex pattern,
        which results in using re.search to match the attribute value.
        By default None, if not provided, default pattern matching any sequence
        of characters is used.

    Notes
    -----
    For more information about attribute selectors see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/Attribute_selectors
    """

    name: str
    value: Optional[PatternType] = None

    def __post_init__(self) -> None:
        """Sets pattern attribute used in SoupSelector find operations."""
        self._pattern = self._parse_pattern()

    def _parse_pattern(self) -> PatternType:
        """Parses pattern used in find methods based on provided init parameters."""
        # if value was not provided, fall back to default pattern
        if self.value is None:
            return re.compile(ns.DEFAULT_PATTERN)
        # cast value to string if not a regex pattern
        if not isinstance(self.value, Pattern):
            return str(self.value)
        # value is already a compiled regex pattern
        return self.value

    def find_all(
        self,
        tag: ns.Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        params = {ns.ATTRS: {self.name: self._pattern}}
        return tag.find_all(**params, recursive=recursive, limit=limit)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AttributeSelector):
            return False

        # pattern is what is used in find methods
        return self._pattern == other._pattern and self.name == other.name


class _SpecificAttributeSelector(AttributeSelector):
    """
    Base class for specific attribute selectors,
    that wraps AttributeSelector with default attribute name for user convenience.

    Child classes should define _NAME attribute with default attribute name,
    that will be used in the AttributeSelector.
    """

    _NAME: str

    def __init__(self, value: Optional[PatternType] = None) -> None:
        """
        Initializes specific attribute selector with default attribute name.

        Parameters
        ----------
        value : str | Pattern, optional
            Value of the attribute, like a class name. Can be regex pattern,
            which results in using re.search to match the attribute value.
            By default None, if not provided, default pattern matching any sequence
            of characters is used.
        """
        super().__init__(name=self._NAME, value=value)


class IdSelector(_SpecificAttributeSelector):
    """
    Component for selecting HTML elements based on 'id' attribute value.
    Convenience wrapper for AttributeSelector with default attribute name 'id',
    as its use is very common in CSS selectors.

    Example
    -------
    >>> IdSelector("main")

    matches all elements that have 'id' attribute with value "main".

    Example
    -------
    >>> <div id="main">Hello World</div> ✔️
    >>> <div id="content">Hello World</div> ❌

    Notes
    -----
    For more information about id attribute see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/ID_selectors
    """

    _NAME = "id"


class ClassSelector(_SpecificAttributeSelector):
    """
    Component for selecting HTML elements based on 'class' attribute value.
    Convenience wrapper for AttributeSelector with default attribute name 'class',
    as its use is very common in CSS selectors.

    Example
    -------
    >>> ClassSelector("main")

    matches all elements that have 'class' attribute with value "main".

    Example
    -------
    >>> <div class="main">Hello World</div> ✔️
    >>> <div class="content">Hello World</div> ❌

    Notes
    -----
    For more information about class attribute see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/Class_selectors
    """

    _NAME = "class"



"""
Module that contains relative selectors and utility components.
"""

from collections.abc import Callable
from typing import Optional

from bs4 import Tag

from soupsavvy.selectors.base import CompositeSoupSelector, SoupSelector, check_selector
from soupsavvy.selectors.tag_utils import TagIterator, TagResultSet


class RelativeSelector(SoupSelector):
    """
    Base class for relative selectors, that are used to find tags relative
    to the tag that is being searched, which is considered an anchor.

    CSS definition of relative selectors states that it is a selector representing
    an element relative to one or more anchor elements preceded by a combinator.

    In this use case, the anchor element is the tag that is being searched, and
    the combinator is the logic of specific relative selector that is used.

    Example
    -------
    >>> selector = Anchor > TypeSelector("div")
    >>> selector.find_all(tag)

    Uses RelativeChild selector to find any div tag that is a direct child of the
    tag that is being searched (passed as an argument).

    In css such selectors can be used in :has pseudo-class, where selector
    is anchored to the element:

    Example
    -------
    >>> div:has(> a, + p)

    Selects any div tag that has a direct child 'a' tag or a next sibling 'p' tag.
    """

    def __init__(self, selector: SoupSelector) -> None:
        """
        Initializes RelativeSelector instance with specified selector.

        Parameters
        ----------
        selector : SoupSelector
            Selector that is used to find tags relative to the anchor tag.
        """
        self._selector = check_selector(selector)

    @property
    def selector(self) -> SoupSelector:
        """
        Returns selector used to find elements relative to the anchor element
        in this relative selector.

        Returns
        -------
        SoupSelector
            Selector used in this relative selector.
        """
        return self._selector

    def __eq__(self, other: object) -> bool:
        # check for RelativeSelector type for type checking sake
        if not isinstance(other, RelativeSelector):
            return False
        elif type(self) is not type(other):
            # checking for exact type match - isinstance(other, self.__class__)
            # when other is subclass of self.__class__ would call other.__eq__(self)
            # which is not desired behavior, as it returns False
            return False

        return self.selector == other.selector

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.selector})"

    def __repr__(self) -> str:
        return str(self)


class BaseRelativeSibling(RelativeSelector):
    """
    Base class with implementation for relative sibling selectors,
    searches for next sibling(s) of the anchor tag.
    Child class needs to define '_limit' class attribute to specify
    how many next siblings to search for and '_func' class attribute
    to specify which method to use for finding siblings.
    """

    # limit of next siblings to check
    _limit: Optional[int]
    # function to use for finding siblings
    _func: Callable[..., list]

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        search = tag.find_parents()[0]
        # find all sibling tags that match the selector
        matching = TagResultSet(
            self.selector.find_all(search, recursive=False),
        )
        siblings = TagResultSet(self.__class__._func(tag, limit=self._limit))
        # find intersection between two sets
        matches = matching & siblings
        return matches.fetch(limit)


class BaseAncestorSelector(RelativeSelector):
    """
    Base class with implementation for ancestor selectors,
    searches for ancestor(s) of the anchor tag.
    Child class needs to define '_limit' class attribute to specify
    how many ancestors to search for.
    """

    _limit: Optional[int]

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        limit = limit or self._limit
        # get max number of ancestors that can possibly be returned
        ancestors = tag.find_parents(limit=self._limit)

        if not ancestors:
            # if no ancestors, make no sense to search
            return []

        search = ancestors[-1].parent or ancestors[-1]
        # search within parent of last ancestor
        matching = TagResultSet(self.selector.find_all(search))
        matches = TagResultSet(ancestors) & matching
        return matches.fetch(limit)


class RelativeChild(RelativeSelector):
    """
    RelativeChild selector is used to find tags matching selector that are
    direct children of the tag that is being searched.

    For RelativeChild selector, with TypeSelector targeting 'p' tag.

    Example
    -------
    >>> RelativeChild(TypeSelector("p"))

    when 'div' element is passed into find methods:

    Example
    -------
    >>> <div><p></p><a></a></div> ✔️
    >>> <div><a><p></p></a></div> ❌
    >>> <div><a></a></div> ❌

    Returns element only if it is a direct child of the anchor element
    and matches the selector.

    RelativeChild selector is equivalent to using '>' combinator in css selectors:

    Example
    -------
    >>> div > p { color: red; }

    It can be created with Anchor instance as well with use of '>' operator:

    Example
    -------
    >>> Anchor > TypeSelector("p")

    Which is equivalent to the first example and returns RelativeChild selector.

    Notes
    -------
    Behavior of RelativeChild selector is equivalent to using find methods of
    selector with recursive=False parameter and is implemented to support
    'has' and ChildCombinator selector.
    """

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        return self.selector.find_all(tag, recursive=False, limit=limit)


class RelativeDescendant(RelativeSelector):
    """
    RelativeDescendant selector is used to find tags matching selector that are
    descendants of the tag that is being searched.

    For RelativeDescendant selector, with TypeSelector targeting 'p' tag,

    Example
    -------
    >>> RelativeDescendant(TypeSelector("p"))

    when 'div' element is passed into find methods:

    Example
    -------
    >>> <div><p></p><a></a></div> ✔️
    >>> <div><a><p></p></a></div> ✔️
    >>> <div><a></a></div> ❌
    >>> <div></div><p></p> ❌

    Returns element only if it is a descendant of the anchor element
    and matches the selector.

    RelativeDescendant selector is equivalent to using whitespace " " combinator
    in css selectors.

    Example
    -------
    >>> div p { color: red; }

    It can be created with Anchor instance as well with use of '>>' operator:

    Example
    -------
    >>> Anchor >> TypeSelector("p")

    Which is equivalent to the first example and returns RelativeDescendant selector.

    Notes
    -------
    Behavior of RelativeDescendant selector is equivalent to using find methods of
    selector with default recursive=True parameter and is implemented to support
    'has' and DescendantCombinator selector.
    """

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        return self.selector.find_all(tag, recursive=True, limit=limit)


class RelativeNextSibling(BaseRelativeSibling):
    """
    RelativeNextSibling selector is used to find tags matching selector that are
    next siblings of the tag that is being searched.

    For RelativeNextSibling selector, with TypeSelector targeting 'p' tag,

    Example
    -------
    >>> RelativeNextSibling(TypeSelector("p"))

    when 'div' element is passed into find methods:

    Example
    -------
    >>> <div></div><p></p> ✔️
    >>> <div></div><a></a><p></p>  ❌
    >>> <p></p><div></div> ❌

    Returns element only if it is a next sibling of the anchor element
    and matches the selector.

    RelativeNextSibling selector is equivalent to using + combinator in css selectors.

    Example
    -------
    >>> div + p { color: red; }

    It can be created with Anchor instance as well with use of '+' operator:

    Example
    -------
    >>> Anchor + TypeSelector("p")

    Which is equivalent to the first example and returns RelativeNextSibling selector.
    """

    _limit = 1
    _func = Tag.find_next_siblings


class RelativeSubsequentSibling(BaseRelativeSibling):
    """
    RelativeSubsequentSibling selector is used to find tags matching selector that are
    subsequent siblings of the tag that is being searched.

    For RelativeSubsequentSibling selector, with TypeSelector targeting 'p' tag,

    Example
    -------
    >>> RelativeSubsequentSibling(TypeSelector("p"))

    when 'div' element is passed into find methods:

    Example
    -------
    >>> <div></div><p></p> ✔️
    >>> <div></div><a></a><p></p> ✔️
    >>> <p></p><div></div> ❌
    >>> <div></div><span><p></p></span> ❌

    Returns element only if it is a subsequent sibling of the anchor element
    and matches the selector.

    RelativeSubsequentSibling selector is equivalent
    to using ~ combinator in css selectors.

    Example
    -------
    >>> div ~ p { color: red; }

    It can be created with Anchor instance as well with use of '*' operator:

    Example
    -------
    >>> Anchor * TypeSelector("p")

    Which is equivalent to the first example
    and returns RelativeSubsequentSibling selector.
    """

    _limit = None
    _func = Tag.find_next_siblings


class RelativeParent(BaseAncestorSelector):
    """
    `RelativeParent` selector is used to find tags matching selector that are
    parent of the tag that is being searched.

    For `RelativeParent` selector, with `TypeSelector` targeting 'div' tag,

    Example
    -------
    >>> RelativeParent(TypeSelector("div"))

    when 'p' element is passed into find methods:

    Example
    -------
    >>> <div><p></p></div> ✔️
    >>> <div><a><p></p></a></div> ❌
    >>> <span><p></p></span> ❌

    Returns element only if it is a parent of the anchor element
    and matches the selector.

    Although this combinator does not have its counterpart in CSS, it can be
    represented as has selector, where child combinator is explicitly stated:

    Example
    -------
    >>> div:has(> p)

    Where 'p' element is an anchor element passed to find methods and 'div' is
    the selector that matches its parent.

    It can be created with Anchor instance as well with use of '<' operator:

    Example
    -------
    >>> Anchor < TypeSelector("div")

    Which is equivalent to the first example and returns `RelativeParent` selector.

    Notes
    -------
    `RelativeParent` selector ignores `recursive` parameter,
    as it is always searches only for parent of the anchor element,
    `find_all` method can return at most one element (parent).
    """

    _limit = 1


class RelativeAncestor(BaseAncestorSelector):
    """
    `RelativeAncestor` selector is used to find tags matching selector that are
    ancestors of the tag that is being searched.

    For `RelativeAncestor` selector, with `TypeSelector` targeting 'div' tag,

    Example
    -------
    >>> RelativeAncestor(TypeSelector("div"))

    when 'p' element is passed into find methods:

    Example
    -------
    >>> <div><p></p></div> ✔️
    >>> <div><a><p></p></a></div> ✔️
    >>> <span><p></p></span> ❌
    >>> <p></p><div></div> ❌

    Returns element only if it is an ancestor of the anchor element
    and matches the selector.

    Although this combinator does not have its counterpart in CSS, it can be
    represented as has selector, where descendant combinator is implied:

    Example
    -------
    >>> div:has(p)

    Where 'p' element is an anchor element passed to find methods and 'div' is
    the selector that is being searched among its ancestors.

    It can be created with Anchor instance as well with use of '<<' operator:

    Example
    -------
    >>> Anchor << TypeSelector("div")

    Which is equivalent to the first example and returns `RelativeAncestor` selector.

    Notes
    -------
    `RelativeAncestor` selector ignores `recursive` parameter,
    as it is always searches among all ancestors of the anchor element.
    """

    _limit = None


class _Anchor:
    """
    Anchor components is a way to create relative selectors in a more
    readable way by using operators.

    _Anchor is an internal class and it's advisable to use the Anchor
    instance instead, which can be considered like a singleton.

    Component supports the following operators:
    * `>`: RelativeChild

    Example
    -------
    >>> Anchor > TypeSelector("div")

    Creates a RelativeChild selector, that selects any div tag that is a direct
    child of the tag that is being searched.

    * `>>`: RelativeDescendant

    Example
    -------
    >>> Anchor >> TypeSelector("div")

    Creates a RelativeDescendant selector, that selects any div tag that is a descendant
    of the tag that is being searched. This is default behavior of selectors,
    and is equivalent to using the TypeSelector directly, but is implemented
    for the sake of consistency.

    * `+`: RelativeNextSibling

    Example
    -------
    >>> Anchor + TypeSelector("div")

    Creates a RelativeNextSibling selector, that selects any div tag that is next
    sibling of the tag that is being searched, it can logically return at most one tag.

    * `*`: RelativeSubsequentSibling

    Example
    -------
    >>> Anchor * TypeSelector("div")

    Creates a RelativeSubsequentSibling selector, that selects any div tag
    that is a subsequent sibling of the tag that is being searched.

    * `<`: RelativeParent

    Example
    -------
    >>> Anchor < TypeSelector("div")

    Creates a RelativeParent selector, that selects any div tag
    that is a parent of the tag that is being searched.

    * `<<`: RelativeAncestor

    Example
    -------
    >>> Anchor << TypeSelector("div")

    Creates a RelativeAncestor selector, that selects any div tag
    that is an ancestor of the tag that is being searched.

    This imitates css selector relative selectors that are used for example in
    :has pseudo-class, that accepts relative selector list as an argument.

    Example
    -------
    >>> :has(> div, + a)

    This translated to soupsavvy would be:

    Example
    -------
    >>> HasSelector(Anchor > TypeSelector("div"), Anchor + TypeSelector("a"))

    Which would match any tag that has a direct child 'div' and a next sibling 'a' tag.
    Selected tag is the anchor tag that is being searched.

    Notes
    -------
    For more information on relative selectors, see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_selectors/Selector_structure#relative_selector
    """

    def __gt__(self, x: SoupSelector) -> RelativeChild:
        return RelativeChild(check_selector(x))

    def __rshift__(self, x: SoupSelector) -> RelativeDescendant:
        return RelativeDescendant(check_selector(x))

    def __lt__(self, x: SoupSelector) -> RelativeParent:
        return RelativeParent(check_selector(x))

    def __lshift__(self, x: SoupSelector) -> RelativeAncestor:
        return RelativeAncestor(check_selector(x))

    def __add__(self, x: SoupSelector) -> RelativeSelector:
        return RelativeNextSibling(check_selector(x))

    def __mul__(self, x: SoupSelector) -> RelativeSelector:
        return RelativeSubsequentSibling(check_selector(x))


# instance of Anchor class with purpose to be used as a singleton
Anchor = _Anchor()


class HasSelector(CompositeSoupSelector):
    """
    Class representing elements selected with respect to matching reference elements.
    Element is selected if any of the provided selectors matched reference element.

    Example
    -------
    >>> HasSelector(TypeSelector("div"))

    matches all elements that have any descendant with "div" tag name.
    It uses default combinator of relative selector, which is descendant.

    Example
    -------
    >>> <span><div>Hello World</div></span> ✔️
    >>> <span><a>Hello World</a></span> ❌

    Other relative selectors can be used with Anchor element.

    Example
    -------
    >>> from soupsavvy.selectors.relative import Anchor
    >>> HasSelector(Anchor > TypeSelector("div"))
    >>> HasSelector(Anchor + TypeSelector("div"))

    or by using RelativeSelector components directly:

    Example
    -------
    >>> from soupsavvy.selectors.relative import RelativeChild, RelativeNextSibling
    >>> HasSelector(RelativeChild(TypeSelector("div")))
    >>> HasSelector(RelativeNextSibling(TypeSelector("div"))

    Example
    -------
    >>> <span><div>Hello World</div></span> ✔️
    >>> <span><a><div>Hello World</div></a></span> ❌

    In this case, HasSelector is anchored against any element, and matches only elements
    that have "div" tag name as a child.

    Object can be initialized with multiple selectors as well, in which case
    at least one selector must match for element to be included in the result.

    This is an equivalent of CSS :has() pseudo-class,
    that represents elements if any of the relative selectors that are passed as an argument
    match at least one element when anchored against this element.

    Example
    -------
    >>> :has(div, a) { color: red; }
    >>> :has(+ div, > a) { color: red; }

    These examples translated to soupsavvy would be:

    Example
    -------
    >>> from soupsavvy.selectors.relative import Anchor
    >>> HasSelector(TypeSelector("div"), TypeSelector("a"))
    >>> HasSelector(Anchor + TypeSelector("div"), Anchor > TypeSelector("a"))

    Notes
    -----
    Passing RelativeDescendant selector into HasSelector is equivalent to using
    its selector directly, as descendant combinator is default option.

    Example
    -------
    >>> HasSelector(RelativeDescendant(TypeSelector("div")))
    >>> HasSelector(Anchor > TypeSelector("div"))
    >>> HasSelector(TypeSelector("div"))

    Three of the above examples are equivalent.

    For more information on :has() pseudo-class see:
    https://developer.mozilla.org/en-US/docs/Web/CSS/:has
    """

    def __init__(
        self,
        selector: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes AndSelector object with provided
        positional arguments as selectors.

        Parameters
        ----------
        selectors: SoupSelector
            SoupSelector objects to match accepted as positional arguments.
            At least one selector is required to create HasSelector.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of SoupSelector.
        """
        super().__init__([selector, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:

        elements = TagIterator(tag, recursive=recursive)
        matching: list[Tag] = []

        for element in elements:
            # we only care if anything matching was found
            if any(step.find(element) for step in self.selectors):
                matching.append(element)

                if len(matching) == limit:
                    break

        return matching



"""Module with utility components operations with css nth-selectors."""

import math
import re
from collections.abc import Callable, Generator
from dataclasses import dataclass
from itertools import count

# matches nth-selector that contains only step parameter - e.g. "2"
OFFSET_PATTERN = re.compile(r"^\d+$")
# matches nth-selector that contains only offset parameter - e.g. "-3n", "n"
STEP_PATTERN = re.compile(r"^-?\d*n$")
# matches nth-selector that contains both step and offset parameters - e.g. "-3n+2"
COMBINED_PATTERN = re.compile(r"^-?\d*n\+\d+$")


@dataclass
class NthGenerator:
    """
    Class for generating a list of integers that match the nth-child CSS selector.

    Parameters
    ----------
    step : int
        The step parameter of the nth-child CSS selector. Must be an integer.
    offset : int
        The offset parameter of the nth-child CSS selector.
        Must be a non-negative integer.

    Examples
    --------
    >>> nth = NthGenerator(step=2, offset=1)
    >>> list(nth.generate(10))
    [1, 3, 5, 7, 9]

    >>> nth = NthGenerator(step=3, offset=2)
    >>> list(nth.generate(10))
    [2, 5, 8]

    Generate method has stop parameter that limits the range of integers generated.
    When having list of 10 sibling elements, the stop parameter can be set to 10
    to determine which elements match provided nth-child selector.

    Examples
    --------
    >>> nth = NthGenerator(step=-1, offset=3)
    >>> elements = ["a", "b", "c", "d", "e", "f"]
    >>> index = [x-1 for x in nth.generate(len(elements))]
    >>> [elements[i] for i in index]
    ["d", "e", "f"]

    This selector matches first three children, and can be used the way shown above
    to filter out the elements that match the selector.
    """

    step: int
    offset: int

    def __post_init__(self):
        """
        Checks if passed parameters are valid.

        Raises
        ------
        TypeError
            If step or offset are not integers.
        ValueError
            If offset is negative.
        """
        if not isinstance(self.step, int):
            raise TypeError(
                f"Step parameter must be an integer, not {type(self.step)}."
            )
        if not isinstance(self.offset, int):
            raise TypeError(
                f"Offset parameter must be an integer, not {type(self.offset)}."
            )
        if self.offset < 0:
            raise ValueError("Offset must not be negative.")

    @property
    def _increasing(self) -> bool:
        """
        Returns true if linear function: y = ax + b where a = step and b = offset
        is increasing (step is positive), false otherwise.
        """
        return self.step > 0

    @property
    def _func(self) -> Callable[[int], int]:
        """
        Returns a linear function that represents the nth-child selector.
        y = ax + b where a = step and b = offset, which is used to generate
        the list of integers that match the selector.
        """
        a, b = self.step, self.offset
        return lambda x: a * x + b

    def generate(self, stop: int) -> Generator[int, None, None]:
        """
        Generates a list of integers that match the nth-child selector.

        Parameters
        ----------
        stop : int
            The stop parameter that limits the range of integers generated.
            Must be a positive integer.

        Yields
        ------
        int
            Integers that match the nth-child selector.
            Generated integers fall into the range from 1 to stop.

        Raises
        ------
        TypeError
            If stop parameter is not an integer.

        Examples
        --------
        >>> nth = NthGenerator(step=2, offset=1)
        >>> list(nth.generate(10))
        [1, 3, 5, 7, 9]

        >>> nth = NthGenerator(step=-3, offset=20)
        >>> list(nth.generate(8))
        [8, 5, 2]
        """
        if not isinstance(stop, int):
            raise TypeError(f"Stop parameter must be an integer, not {type(stop)}.")

        if stop < 1:
            return

        a, b = self.step, self.offset

        if a == 0:
            if 1 <= b <= stop:
                yield b
            return

        c = ((1 - b) / a) if self._increasing else ((stop - b) / a)
        start = max(math.ceil(c), 0)

        for x in count(start):
            y = self._func(x)

            if not (1 <= y <= stop):
                break

            yield y


def parse_nth(selector: str) -> NthGenerator:
    """
    Parse CSS nth selector and return NthGenerator instance.

    Parameters
    ----------
    selector : str
        CSS nth selector string. Accepts all valid css nth selectors.

    Returns
    -------
    NthGenerator
        NthGenerator instance that represents the parsed nth selector.

    Raises
    ------
    ValueError
        If the selector is invalid.

    Handles "odd" and "even" literals and any number of whitespace characters
    in the selector string.

    Examples
    --------
    >>> nth = parse_nth("2")
    NthGenerator(step=0, offset=2)

    >>> nth = parse_nth("-3n+2")
    NthGenerator(step=-3, offset=2)

    >>> nth = parse_nth("odd")
    NthGenerator(step=2, offset=1)

    >>> nth = parse_nth("15n")
    NthGenerator(step=15, offset=0)
    """
    # handle whitespaces
    selector = selector.replace(" ", "")

    # handle special cases
    if selector == "odd":
        return NthGenerator(2, 1)
    if selector == "even":
        return NthGenerator(2, 0)

    # handle non-step cases
    if OFFSET_PATTERN.match(selector):
        return NthGenerator(0, int(selector))

    pattern_match = STEP_PATTERN.match(selector) or COMBINED_PATTERN.match(selector)

    if not pattern_match:
        raise ValueError(f"Invalid CSS nth selector: {selector}")

    # replaces n with no preceding digit with 1n - e.g. "n" -> "1n", "-n" -> "-1n"
    selector = re.sub(r"(\D|^)n", r"\g<1>1n", selector)
    # replaces missing offset with 0 - e.g. "2n" -> "2n0"
    selector = re.sub(r"n$", "n0", selector)

    step, offset = map(int, selector.split("n"))
    return NthGenerator(step, offset)



"""
Module with base model class used as a parent for all user defined
`soupsavvy` models.
"""

from __future__ import annotations

from abc import ABC
from functools import reduce
from typing import Any, Literal, Optional, Type, TypeVar, overload

from bs4 import Tag
from typing_extensions import Self

import soupsavvy.exceptions as exc
import soupsavvy.models.constants as c
from soupsavvy.interfaces import Comparable, TagSearcher, TagSearcherExceptions
from soupsavvy.selectors.base import SoupSelector, check_selector

T = TypeVar("T")


class ModelMeta(type(ABC)):
    """
    Metaclass for all models derived from `BaseModel`. This metaclass ensures that
    certain attributes and methods are defined
    and properly configured in each model class.

    It handles validation of provided attributes and controls inheritance of fields.
    Meta inherits from `type(ABC)` to avoid metaclass conflicts.

    Attributes
    ----------
    scope : SoupSelector
        Returns the `__scope__` attribute, which defines the scope
        within which the model is to be found.
    fields : dict[str, TagSearcher]
        Fields that defines the model and search operations.
    """

    def __init__(cls, name, bases, class_dict):
        """
        Initializes the model class and validates its attributes.
        For each user-defined model, which is a subclass of `BaseModel`,
        checks if scope and fields are properly defined.

        Raises
        ------
        ScopeNotDefinedException
            If the scope attribute is missing in the model class.
        FieldsNotDefinedException
            If no fields are defined in the model class.
        """
        super().__init__(name, bases, class_dict)

        if name in c.BASE_MODELS:
            return

        scope = getattr(cls, c.SCOPE)

        if scope is None:
            raise exc.ScopeNotDefinedException(
                f"Missing '{c.SCOPE}' class attribute in '{name}'."
            )

        message = (
            f"'{c.SCOPE}' attribute of the model class '{cls.__name__}' "
            f"needs to be '{SoupSelector.__name__}' instance, got '{type(scope)}'."
        )
        check_selector(scope, message=message)

        if not cls._get_fields():
            raise exc.FieldsNotDefinedException(
                f"Model '{cls.__name__}' has no fields defined. At least one required."
            )

    @property
    def scope(cls) -> SoupSelector:
        """
        Returns the `__scope__` attribute,
        which defines the scope selector for the model.

        Returns
        -------
        SoupSelector
            The scope selector used to identify the element in which
            the model is searched.
        """
        return getattr(cls, c.SCOPE)

    @property
    def fields(cls) -> dict[str, TagSearcher]:
        """
        Returns the fields of the model class with their
        respective TagSearcher instances. The fields are aggregated based on
        the `__inherit_fields__` setting.

        Returns
        -------
        dict[str, TagSearcher]
            A dictionary mapping field names to their respective TagSearcher instances.
        """
        return cls._get_fields()

    def _get_fields(cls) -> dict[str, TagSearcher]:
        """
        Returns the fields of the model class with their
        respective TagSearcher instances based on the `__inherit_fields__` setting.
        """
        classes = (
            [
                base
                for base in reversed(cls.__mro__)
                if issubclass(base, BaseModel) and base is not BaseModel
            ]
            if getattr(cls, c.INHERIT_FIELDS)
            else [cls]
        )

        return reduce(
            dict.__or__,
            [
                {
                    key: value
                    for key, value in class_.__dict__.items()
                    if (
                        # accepted field must be TagSearcher
                        isinstance(value, TagSearcher)
                        # or BaseModel subclass
                        or (isinstance(value, type) and issubclass(value, BaseModel))
                    )
                    and key not in c.SPECIAL_FIELDS
                }
                for class_ in classes
            ],
        )


class BaseModel(TagSearcher, Comparable, metaclass=ModelMeta):
    """Base class for all user-defined models in `soupsavvy`."""

    __scope__: SoupSelector = None  # type: ignore
    __inherit_fields__: bool = True

    def __init__(self, **kwargs) -> None:
        """
        Initializes a model instance with provided field values.
        Model should not be initialized directly, but through the `find` methods.

        Parameters
        ----------
        kwargs : Any
            Field values to initialize the model with provided
            as keyword arguments.

        Raises
        ------
        MissingFieldsException
            If any required fields are missing from the provided kwargs.
        UnknownModelFieldException
            If any unknown fields are provided in the kwargs.
        """
        fields = self.__class__.fields.keys()

        for field in fields:

            try:
                value = kwargs.pop(field)
            except KeyError:
                raise exc.MissingFieldsException(
                    f"Cannot initialize model '{self.__class__.__name__}' "
                    f"without '{field}' field."
                )

            setattr(self, field, value)

        if kwargs:
            raise exc.UnknownModelFieldException(
                f"Model '{self.__class__.__name__}' has no fields defined "
                f"for: {kwargs.keys()}"
            )

        self.__post_init__()

    def __post_init__(self) -> None:
        """
        Post-initialization hook. Can be overridden by subclasses to perform additional
        initialization steps after the base initialization.
        """

    @property
    def attributes(self) -> dict[str, Any]:
        """
        Returns a dictionary of model instance attributes representing
        model fields and their respective values.

        Returns
        -------
        dict[str, Any]
            A dictionary mapping model field names to their respective values.
        """
        return {field: getattr(self, field) for field in self.__class__.fields.keys()}

    @overload
    @classmethod
    def find(
        cls,
        tag: Tag,
        strict: Literal[True] = ...,
        recursive: bool = ...,
    ) -> Self: ...

    @overload
    @classmethod
    def find(
        cls,
        tag: Tag,
        strict: Literal[False] = ...,
        recursive: bool = ...,
    ) -> Optional[Self]: ...

    @overload
    @classmethod
    def find(
        cls,
        tag: Tag,
        strict: bool = ...,
        recursive: bool = ...,
    ) -> Optional[Self]: ...

    @classmethod
    def find(
        cls,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Optional[Self]:
        """
        Searches for and returns an instance of the model within the provided tag.
        By default, perform recursive, non-strict search for model fields
        within the scope element.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object within which to search for the model.
        strict : bool, optional
            Whether to raise an exception if the scope element
            for the model is not found. Default is False, None is returned.
        recursive : bool, optional
            Whether the search for the model scope element should be recursive.
            Default is True.

        Returns
        -------
        Self | None
            An instance of the model if found, otherwise None.

        Raises
        ------
        ModelNotFoundException
            If the model's scope is not found and strict is True.
        FieldExtractionException
            If any model field failed to be extracted.
        """
        bound = cls.scope.find(tag=tag, strict=False, recursive=recursive)

        if bound is not None:
            return cls._find(bound)

        if strict:
            raise exc.ModelNotFoundException(
                f"Scope for the model '{cls.__name__}' was not found "
                f"from '{cls.scope}' in tag."
            )

        return None

    @classmethod
    def _find(cls, tag: Tag) -> Self:
        """
        Internal method to find and initialize a model instance from a given tag.
        By default, perform recursive, non-strict search for model fields.

        Parameters
        ----------
        tag : Tag
            The tag within which to search for model fields (scope element).

        Returns
        -------
        Self
            An instance of the model.

        Raises
        ------
        FieldExtractionException
            If any model field failed to be extracted.
        """
        params = {}

        for key, selector in cls.fields.items():
            try:
                result = selector.find(
                    tag=tag,
                    strict=c.DEFAULT_STRICT,
                    recursive=c.DEFAULT_RECURSIVE,
                )
            except exc.RequiredConstraintException as e:
                raise exc.FieldExtractionException(
                    f"Field '{key}' is required and was not found in model '{cls.__name__}' "
                    f"for element:\n{tag}."
                ) from e
            except TagSearcherExceptions as e:
                raise exc.FieldExtractionException(
                    f"Extracting field '{key}' failed in model '{cls.__name__}' "
                    f"for element:\n{tag}."
                ) from e

            params[key] = result

        return cls(**params)

    @classmethod
    def find_all(
        cls,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Self]:
        """
        Searches for and returns all instances of the model within the provided tag.
        By default, perform recursive, non-strict search for model fields
        just like in `find` method.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object within which to search for the model.
        recursive : bool, optional
            Whether the search for the model scope element should be recursive.
            Default is True.
        limit : int, optional
            Maximum number of model instances to return. Default is None, which
            returns all instances found.

        Returns
        -------
        list[Self]
            A list of model instances found within the tag.
        """
        elements = cls.scope.find_all(tag=tag, recursive=recursive, limit=limit)
        return [cls._find(element) for element in elements]

    def migrate(self, model: Type[T], **kwargs) -> T:
        """
        Migrates the model instance to another model class using its fields
        in target class initialization.

        Parameters
        ----------
        model : Type[Model]
            The target model class to migrate the instance to.
        kwargs : Any
            Additional keyword arguments to pass to model initialization.

        Returns
        -------
        Model
            An instance of the target model class.
        """
        return model(**self.attributes, **kwargs)

    def __str__(self) -> str:
        params = [
            f"{name}={repr(getattr(self, name))}"
            for name in self.__class__.fields.keys()
        ]
        return f"{self.__class__.__name__}({', '.join(params)})"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, x: Any) -> bool:
        """
        Checks if provided object is equal to the model instance.
        They need to be of the same class and have the same field values.
        """
        if not isinstance(x, self.__class__):
            return False

        fields = self.__class__.fields.keys()
        return all(getattr(self, key) == getattr(x, key) for key in fields)



"""
Module for selection pipeline class.
Pipeline for chaining selector and operation together, used as a bridge between
selecting html elements and processing the data.
"""

from __future__ import annotations

from typing import Any, Optional

from bs4 import Tag

import soupsavvy.exceptions as exc
from soupsavvy.interfaces import Comparable, TagSearcher
from soupsavvy.operations.base import BaseOperation, check_operation


class SelectionPipeline(TagSearcher, Comparable):
    """
    Class for chaining searcher and operation together.
    Uses searcher to find information in tag and operation to process the data.

    Example
    -------
    >>> from soupsavvy import TypeSelector
    ... from soupsavvy.operations import Operation, Text
    ... pipeline = TypeSelector("span") | Text()
    ... pipeline.find(soup)
    'information'

    Most common way of creating a pipeline is using the `|` operator
    on selector and operation.
    """

    def __init__(self, selector: TagSearcher, operation: BaseOperation) -> None:
        """
        Initializes SelectionPipeline with selector and operation.

        Parameters
        ----------
        selector : TagSearcher
            Selector used for finding target information in the tag.
        operation : BaseOperation
            Operation used for processing the data.
        """
        if not isinstance(selector, TagSearcher):
            raise exc.NotTagSearcherException(
                f"Expected {TagSearcher.__name__}, got {type(selector)}"
            )

        self._selector = selector
        self._operation = operation

    @property
    def selector(self) -> TagSearcher:
        """
        Returns TagSearcher object of this pipeline used for finding target
        information in the tag.

        Returns
        -------
        TagSearcher
            TagSearcher object used in this pipeline.
        """
        return self._selector

    @property
    def operation(self) -> BaseOperation:
        """
        Returns BaseOperation object of this pipeline used for processing the data.

        Returns
        -------
        BaseOperation
            BaseOperation object used in this pipeline.
        """
        return self._operation

    def find(
        self,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Any:
        """
        Finds a first element matching selector and processes it with operation.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object.
        strict : bool, optional
            If True, raises an exception if tag was not found in markup,
            if False and tag was not found, returns None by defaulting to BeautifulSoup
            implementation. Value of this parameter does not affect behavior if tag
            was successfully found in the markup. By default False.
        recursive : bool, optional
            bs4.Tag.find method parameter that specifies if search should be recursive.
            If set to False, only direct children of the tag will be searched.
            By default True.

        Returns
        -------
        Any
            Result of the operation applied to the found tag.

        Raises
        ------
        TagNotFoundException
            If strict parameter is set to True and tag was not found in markup.
        """
        return self.operation.execute(
            self.selector.find(tag, strict=strict, recursive=recursive)
        )

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Any]:
        """
        Finds all elements matching selector and processes them with operation.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup Tag object.
        recursive : bool, optional
            bs4.Tag.find method parameter that specifies if search should be recursive.
            If set to False, only direct children of the tag will be searched.
            By default True.
        limit : int, optional
            bs4.Tag.find_all method parameter that specifies maximum number of results
            to return. If set to None, all results are returned. By default None.

        Returns
        -------
        list[Any]
            A list of results, if none found, the list is empty.
        """
        return [
            self.operation.execute(element)
            for element in self.selector.find_all(tag, recursive=recursive, limit=limit)
        ]

    def __or__(self, x: Any) -> SelectionPipeline:
        """
        Overrides __or__ method called also by pipe operator '|'.
        Creates new SelectionPipeline by extending operations with provided one.

        Parameters
        ----------
        x : BaseOperation
            BaseOperation object used to extend the pipeline.

        Returns
        -------
        SelectionPipeline
            New `SelectionPipeline` with extended operations.

        Raises
        ------
        NotOperationException
            If provided object is not of type BaseOperation.
        """
        x = check_operation(x)
        operation = self.operation | x
        return SelectionPipeline(selector=self.selector, operation=operation)

    def __eq__(self, x) -> bool:
        # equal only if both selector and operation are the same
        if not isinstance(x, SelectionPipeline):
            return False

        return self.selector == x.selector and self.operation == x.operation



"""Namespace with constants and types for tag components."""

from typing import Pattern, Union

from bs4 import NavigableString, Tag

# bs4 parameter name for tag name
NAME = "name"
# attribute filter parameter name in bs4.Tag find methods
ATTRS = "attrs"
# bs4 parameter name for tag text
STRING = "string"
# possible return types of bs4.Tag find method
FindResult = Union[NavigableString, Tag, None]
# default pattern if no value was provided
# wild card - matches any sequence of characters in a non-greedy
DEFAULT_PATTERN = "(.*?)"
# css selector wildcard
CSS_SELECTOR_WILDCARD = "*"
# allowed pattern types for attribute selectors
PatternType = Union[Pattern[str], str]



"""
Module with wrappers for operations and selectors used mostly in models.
Wrappers are used to control behavior and define how fields of the models are defined.

Classes
-------
- SkipNone : OperationWrapper
- Suppress : OperationWrapper
- All : FieldWrapper
- Required : FieldWrapper
- Default : FieldWrapper
"""

from typing import Any, Optional, Type, Union

from bs4 import Tag

import soupsavvy.exceptions as exc
from soupsavvy.interfaces import Comparable, TagSearcher
from soupsavvy.operations.base import BaseOperation, check_operation
from soupsavvy.operations.selection_pipeline import SelectionPipeline

# allowed exception types to suppress
ExceptionCategory = Union[Type[Exception], tuple[Type[Exception], ...]]


class OperationWrapper(BaseOperation):
    """
    A wrapper class for operations that extends the base operation functionality.
    Acts as a higher order operation, which controls behavior of the wrapped operation.
    """

    def __init__(self, operation: BaseOperation) -> None:
        """
        Initialize Higher Order Operation.

        Parameters
        ----------
        operation : BaseOperation
            The operation to be wrapped.

        Raises
        ------
        NotOperationException
            If provided object is not an instance of BaseOperation.
        """
        self.operation = check_operation(operation)

    def __eq__(self, x: Any) -> bool:
        """
        Check if two OperationWrapper instances are equal.
        They need to be of the same class and wrap the same operation.

        Parameters
        ----------
        x : Any
            The object to compare with.

        Returns
        -------
        bool
            True if the instances are equal, False otherwise.
        """
        return isinstance(x, self.__class__) and self.operation == x.operation

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.operation})"

    def __repr__(self) -> str:
        return str(self)


class SkipNone(OperationWrapper):
    """
    A wrapper that skips the operation if the input is None.
    Used to prevent exceptions where it's safe and expected to skip the operation.

    Example
    -------
    >>> from soupsavvy.operations import Text
    ... from soupsavvy.models import SkipNone
    ... operation = SkipNone(Text())
    ... operation.execute(None)
    None

    When element was not found, which can be expected, skips operation
    and returns None.
    """

    def _execute(self, arg: Any) -> Any:
        if arg is None:
            return None

        return self.operation.execute(arg)


class Suppress(OperationWrapper):
    """
    A wrapper that executes the operation and suppresses exceptions raised,
    returning None instead. Used to catch exceptions where it's expected
    this might happen.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... from soupsavvy.models import Suppress
    ... operation = Suppress(Operation(int))
    ... operation.execute("")
    None

    This can be used with `Default` operation to provide a default value
    when the operation fails.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... from soupsavvy.models import Suppress
    ... operation = Default(Suppress(Operation(int)), 0)
    ... operation.execute("")
    None

    Operations in example can be used to try to convert string to integer
    from text of tag, that can potentially be empty. In such case, if it's not required,
    default can be set to None or known value.

    Suppress also accepts category of exceptions to suppress, by default it suppresses
    all exceptions that inherit from Exception.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... from soupsavvy.models import Suppress
    ... operation = Suppress(Operation(int), category=ValueError)
    ... operation.execute("not an integer")
    None

    Category can be a tuple of exceptions as well, `issubclass` is used to check
    if the cause of exception is subclass of provided category.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... from soupsavvy.models import Suppress
    ... operation = Suppress(Operation(int), category=(AttributeError, ValueError))
    ... operation.execute("not an integer")
    FailedOperationExecution
    """

    def __init__(
        self,
        operation: BaseOperation,
        category: ExceptionCategory = Exception,
    ) -> None:
        """
        Initialize Suppress OperationWrapper.

        Parameters
        ----------
        operation : BaseOperation
            The operation to be wrapped.
        category : Type[Exception] | tuple[Type[Exception], ...], optional
            The exception type(s) to suppress. By default, suppresses all exceptions
            that inherit from Exception.

        Raises
        ------
        NotOperationException
            If provided object is not an instance of BaseOperation.
        """
        super().__init__(operation)
        self._category = category

    def _execute(self, arg: Any) -> Any:
        try:
            return self.operation.execute(arg)
        except exc.FailedOperationExecution as e:
            cause = e.__cause__.__class__

            if not issubclass(cause, self._category):
                raise
            return None


class FieldWrapper(TagSearcher, Comparable):
    """
    A wrapper for TagSearcher objects, that acts as a higher order searcher,
    which controls behavior of the wrapped searcher.
    Used as field to defined model.
    """

    def __init__(self, selector: TagSearcher) -> None:
        """
        Initialize FieldWrapper.

        Parameters
        ----------
        selector : TagSearcher
            The TagSearcher instance to be wrapped.

        Raises
        ------
        NotTagSearcherException
            If provided object is not an instance of TagSearcher.
        """

        if not isinstance(selector, TagSearcher):
            raise exc.NotTagSearcherException(
                f"Expected {TagSearcher.__name__}, got {type(selector)}."
            )

        self._selector = selector

    @property
    def selector(self) -> TagSearcher:
        """
        Return searcher, that is wrapped by this FieldWrapper.

        Returns
        -------
        TagSearcher
            TagSearcher instance wrapped by this FieldWrapper.
        """
        return self._selector

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Any]:
        """
        Find all matching tags using the wrapped selector.
        Used for compatibility with TagSearcher interface,
        delegates to wrapped selector.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup tag to search within.
        recursive : bool, optional
            Whether to search recursively, by default True.
        limit : int, optional
            Limit the number of results, by default None.

        Returns
        -------
        list[Any]
            A list of found results.
        """
        return self.selector.find_all(tag, recursive=recursive, limit=limit)

    def __eq__(self, x: Any) -> bool:
        """
        Check if two FieldWrapper instances are equal.
        They need to be of the same class and wrap the same selector.

        Parameters
        ----------
        x : Any
            The object to compare with.

        Returns
        -------
        bool
            True if the instances are equal, False otherwise.
        """
        return isinstance(x, self.__class__) and self.selector == x.selector

    def __or__(self, x: Any) -> SelectionPipeline:
        """
        Overrides __or__ method called also by pipe operator '|'.
        Combines the current FieldWrapper with an operation.

        Parameters
        ----------
        x : BaseOperation
            The operation to be combined with the FieldWrapper.

        Returns
        -------
        SelectionPipeline
            New SelectionPipeline object created from combining selector and operation.

        Raises
        ------
        NotOperationException
            If provided object is not of BaseOperation type.
        """
        x = check_operation(x)
        return SelectionPipeline(selector=self, operation=x)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.selector})"

    def __repr__(self) -> str:
        return str(self)


class All(FieldWrapper):
    """
    Field wrapper for selecting multiple elements matching the selector.
    Forces find method to fall back to find_all method and return all matches.

    Example
    -------
    >>> from soupsavvy.models import All
    ... from soupsavvy import TypeSelector
    ... selector = All(TypeSelector("div"))
    ... selector.find(tag)
    [element1, element2, element3]
    """

    def find(self, tag: Tag, strict: bool = False, recursive: bool = True) -> list[Any]:
        """
        Find all matching tags using the wrapped selector,
        enforcing the use of find_all method.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup tag to search within.
        strict : bool, optional
            Ignored, as this method always falls back to find_all.
        recursive : bool, optional
            Whether to search recursively, by default True.

        Returns
        -------
        list[Any]
            A list of matching results.
        """
        return self.selector.find_all(tag, recursive=recursive)


class Required(FieldWrapper):
    """
    Field wrapper for enforcing matched element not to be None.
    Raises an exception if searcher does not find any matches.

    Example
    -------
    >>> from soupsavvy.models import Required
    ... from soupsavvy import TypeSelector
    ... selector = Required(TypeSelector("div"))
    ... selector.find(tag)
    RequiredConstraintException
    """

    def find(self, tag: Tag, strict: bool = False, recursive: bool = True) -> Any:
        """
        Finds a required element using the wrapped selector,
        enforcing matched element not to be None.
        If any exception is raised during the search, it's propagated to the caller.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup tag to search within.
        strict : bool, optional
            If True, raises an exception if no matches are found, by default False.
        recursive : bool, optional
            Whether to search recursively, by default True.

        Returns
        -------
        Any
            The found element.

        Raises
        ------
        RequiredConstraintException
            If selector returns None, indicating that required element was not found.
        """
        result = self.selector.find(tag=tag, strict=strict, recursive=recursive)

        if result is None:
            raise exc.RequiredConstraintException("Required element was not found")

        return result


class Default(FieldWrapper):
    """
    Field wrapper for returning a default value if no match is found.

    Example
    -------
    >>> from soupsavvy.models import Default
    ... from soupsavvy import TypeSelector
    ... selector = Default(TypeSelector("div"), default="1234")
    ... selector.find(tag)
    "1234"
    """

    def __init__(self, selector: TagSearcher, default: Any) -> None:
        """
        Initializes Default FieldWrapper.

        Parameters
        ----------
        selector : TagSearcher
            The TagSearcher instance to be wrapped.
        default : Any
            The default value to return if no match is found.
        """
        super().__init__(selector)
        self.default = default

    def find(self, tag: Tag, strict: bool = False, recursive: bool = True):
        """
        Finds an element, returning a default value if None was returned
        by wrapped selector. Any exception raised during the search is propagated.

        Parameters
        ----------
        tag : Tag
            Any BeautifulSoup tag to search within.
        strict : bool, optional
            If True, raises an exception if no matches are found, by default False.
        recursive : bool, optional
            Whether to search recursively, by default True.

        Returns
        -------
        Any
            The found element or the default value if not found.
        """
        result = self.selector.find(tag=tag, strict=strict, recursive=recursive)
        return self.default if result is None else result

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.selector}, default={self.default})"



"""
Module for general purpose operations.

* Operation - Custom operation with any function.
* OperationPipeline - Chain multiple operations together.
* Text - Extract text from a BeautifulSoup Tag - most common operation.
* Href - Extract href attribute from a BeautifulSoup Tag.
* Parent - Extract parent of a BeautifulSoup Tag.

These components are design to be used for processing html tags and extracting
desired information. They can be used in combination with selectors.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import Any, Optional

from bs4 import Tag

from soupsavvy.operations.base import (
    BaseOperation,
    OperationSearcherMixin,
    check_operation,
)
from soupsavvy.selectors.base import SoupSelector


class OperationPipeline(OperationSearcherMixin):
    """
    Pipeline for chaining multiple operations together.
    Applies each operation in sequence, passing the result to the next.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... pipeline = Operation(int) | Operation(lambda x: x + 1)
    ... pipeline.execute("1")
    2

    Most common way of creating a pipeline is using the `|` operator
    on two operations.

    `OperationPipeline` is operation-searcher mixin, which means it can be used
    to find information in bs4 Tag directly with find methods.
    This way, it can be used as field in model or `execute` method can be replaced
    with `find` method, which would produce the same result.
    """

    def __init__(
        self,
        operation1: BaseOperation,
        operation2: BaseOperation,
        /,
        *operations: BaseOperation,
    ) -> None:
        """
        Initializes OperationPipeline with multiple operations.

        Parameters
        ----------
        operations : BaseOperation
            BaseOperation instances to be chained together.

        Raises
        ------
        NotOperationException
            If any of the input operations is not a valid operation.
        """
        self.operations = [
            check_operation(operation)
            for operation in (operation1, operation2, *operations)
        ]

    def _execute(self, arg: Any) -> Any:
        """
        Executes each operation in sequence, passing the result to the next one.
        """
        for operation in self.operations:
            arg = operation.execute(arg)
        return arg

    def __or__(self, x: Any) -> OperationPipeline:
        """
        Overrides __or__ method called also by pipe operator '|'.
        Creates new `OperationPipeline` with additional provided operation.

        Parameters
        ----------
        x : BaseOperation
            BaseOperation object used to extend the pipeline.

        Returns
        -------
        OperationPipeline
            New `OperationPipeline` with extended operations.

        Raises
        ------
        NotOperationException
            If provided object is not of type BaseOperation.
        """
        x = check_operation(x)
        return OperationPipeline(*self.operations, x)

    def __eq__(self, x: Any) -> bool:
        # equal only if operations are the same
        if not isinstance(x, OperationPipeline):
            return False

        return self.operations == x.operations


class Operation(OperationSearcherMixin):
    """
    Custom operation that wraps any function
    to be used with other `soupsavvy` components.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... operation = Operation(str.lower)
    ... operation.execute("TEXT")
    "text"

    `Operation` is operation-searcher mixin, which means it can be used
    to find information in bs4 Tag directly with find methods.
    This way, it can be used as field in model or `execute` method can be replaced
    with `find` method, which would produce the same result.
    """

    def __init__(self, func: Callable) -> None:
        """
        Initializes Operation with a custom function.

        Parameters
        ----------
        func : Callable
            Any callable object that can be called with one positional argument.
        """
        self.operation = func

    def _execute(self, arg: Any) -> Any:
        """Executes the custom operation."""
        return self.operation(arg)

    def __eq__(self, x: Any) -> bool:
        # functions used as operations need to be the same object
        if not isinstance(x, Operation):
            return False

        return self.operation is x.operation


class Text(OperationSearcherMixin):
    """
    Operation to extract text from a BeautifulSoup Tag.
    Wrapper of most common operation used in web scraping.

    Example
    -------
    >>> from soupsavvy.operations import Text
    ... operation = Text()
    ... operation.execute(tag)
    "Extracted text from the tag"

    Wrapper for BeautifulSoup `get_text` method, that exposes all its options,
    and imitates its default behavior.

    Implements `TagSearcher` interface for convenience. It has find methods
    that can be used to extract text from provided tag.

    Example
    -------
    >>> from soupsavvy.operations import Text
    ... operation = Text()
    ... operation.find(tag)
    "Text"
    """

    def __init__(self, separator: str = "", strip: bool = False) -> None:
        """
        Initializes Text operation with optional separator and strip option.

        Parameters
        ----------
        separator : str, optional
            Separator used to join strings from multiple text nodes, by default "".
        strip : bool, optional
            Flag to strip the text nodes from whitespaces and newline characters,
            by default False.
        """
        self.separator = separator
        self.strip = strip

    def _execute(self, arg: Tag) -> str:
        """Extracts text from a BeautifulSoup Tag."""
        return arg.get_text(separator=self.separator, strip=self.strip)

    def __eq__(self, x: Any) -> bool:
        # equal only if separator and strip are the same
        if not isinstance(x, Text):
            return False

        return self.separator == x.separator and self.strip == x.strip


class Href(OperationSearcherMixin):
    """
    Operation to extract href attribute from a BeautifulSoup Tag.
    Wrapper of one of the common operation used in web scraping.
    If href attribute is not present, returns None.

    Example
    -------
    >>> from soupsavvy.operations import Href
    ... operation = Href()
    ... operation.execute(tag)
    "www.example.com"

    Implements `TagSearcher` interface for convenience. It has find methods
    that can be used to extract href from provided tag.

    Example
    -------
    >>> from soupsavvy.operations import Href
    ... operation = Href()
    ... operation.find(tag)
    "www.example.com"
    """

    def _execute(self, arg: Tag) -> Optional[str]:
        """
        Extracts href attribute from a BeautifulSoup Tag.
        If attribute is not present, returns None.
        """
        return arg.get_attribute_list("href")[0]

    def __eq__(self, x: Any) -> bool:
        # equal only if both are instances of Href
        return isinstance(x, Href)


class Parent(BaseOperation, SoupSelector):
    """
    Operation to extract parent of a BeautifulSoup Tag.

    Example
    -------
    >>> from soupsavvy.operations import Parent
    ... operation = Parent()
    ... operation.execute(tag)
    "<div>...</div>"

    Implements `SoupSelector` interface for convenience and can be used
    to extract parent of a provided tag without any conditions.

    Example
    -------
    >>> from soupsavvy.operations import Parent
    ... operation = Parent()
    ... operation.find(tag)
    "<div>--tag--</div>"

    `Parent` has `BaseOperation` higher in MRO than `SoupSelector`, so, using pipe
    operator `|` on `Parent` object will result in `OperationPipeline` instance.

    Example
    -------
    >>> from soupsavvy.operations import Parent
    ... operation = Parent() | Parent()
    ... operation.execute(tag)
    "<div><div>--tag--</div></div>"

    If element does not have parent, returns None.
    """

    def _execute(self, arg: Tag) -> Optional[Tag]:
        """Extracts parent of a BeautifulSoup Tag."""
        return arg.parent

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        return [self.execute(tag)]

    def __eq__(self, x: Any) -> bool:
        # equal only if both are instances of Parent
        return isinstance(x, Parent)



"""
`soupsavvy` is flexible search engine for `BeautifulSoup`,
designed to provide more powerful capabilities, making more complex searches simple
and web scraping tasks more efficient and manageable.
`soupsavvy` introduces the concept of a `Selector`,
a declarative search procedure designed with simple and readable syntax.
It encapsulates search logic, making it reusable across different scenarios.
The package offers various types of selectors that can be easily combined
to perform more complex searches.
"""

from .selectors import (
    AncestorCombinator,
    Anchor,
    AndSelector,
    AnyTagSelector,
    AttributeSelector,
    ChildCombinator,
    ClassSelector,
    DescendantCombinator,
    HasSelector,
    IdSelector,
    NextSiblingCombinator,
    NotSelector,
    OrSelector,
    ParentCombinator,
    PatternSelector,
    SelectorList,
    SelfSelector,
    SubsequentSiblingCombinator,
    TypeSelector,
    UniversalSelector,
    XORSelector,
)

__version__ = "0.2.3"
__author__ = "sewcio543"

__all__ = [
    "UniversalSelector",
    "AttributeSelector",
    "TypeSelector",
    "PatternSelector",
    "SelectorList",
    "DescendantCombinator",
    "AndSelector",
    "NotSelector",
    "SelfSelector",
    "ChildCombinator",
    "NextSiblingCombinator",
    "SubsequentSiblingCombinator",
    "AncestorCombinator",
    "ParentCombinator",
    "HasSelector",
    "Anchor",
    "OrSelector",
    "ClassSelector",
    "IdSelector",
    "XORSelector",
    "AnyTagSelector",
]



"""
Module for the BaseTemplate class.

Templates are the components used to fill the generated structures with content.
They are the building blocks for the generators.
This way generated content can be controlled and customized.
"""

from abc import abstractmethod

from soupsavvy.testing.generators.base import BaseGenerator


class BaseTemplate(BaseGenerator):
    """
    Abstract base class for templates.

    BaseTemplate is also the subclass of BaseGenerator, all Templates
    implement BaseGenerator interface and are easy to interchange with
    other generators.
    """

    @abstractmethod
    def generate(self) -> str:
        """
        Generates a string based on the template.

        Returns
        -------
        str
            The generated string.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} is a base class and does not implement "
            "'generate' method."
        )



from .selectors import (
    CSS,
    Empty,
    FirstChild,
    FirstOfType,
    LastChild,
    LastOfType,
    NthChild,
    NthLastChild,
    NthLastOfType,
    NthOfType,
    OnlyChild,
    OnlyOfType,
)

__all__ = [
    "CSS",
    "Empty",
    "FirstChild",
    "FirstOfType",
    "LastChild",
    "LastOfType",
    "NthChild",
    "NthLastChild",
    "NthLastOfType",
    "NthOfType",
    "OnlyChild",
    "OnlyOfType",
]
